"""
Test suite for the dependency injection system.
"""

import pytest
from unittest.mock import Mock, patch
from knwl.di import (
    service,
    singleton_service,
    inject_config,
    defaults,
    inject_services,
    auto_inject,
    ServiceProvider,
    container,
)

pytestmark = pytest.mark.basic


class TestDependencyInjection:
    """
    Test cases for the dependency injection decorators and functionality.
    """

    def setup_method(self):
        """Clean up DI container before each test."""
        container._service_registry.clear()
        container._config_registry.clear()
        container._defaults_registry.clear()

    def test_fail_to_inject_service(self):
        """Test behavior when service creation fails."""
        with patch("knwl.di.services") as mock_services:
            mock_services.create_service.side_effect = Exception("Service not found")

            @service("non_existent_service")
            def test_function(svc=None):
                return svc

            with pytest.raises(Exception, match="Service not found"):
                test_function()

            mock_services.create_service.assert_called_once_with(
                "non_existent_service", variant_name=None, override=None
            )

    def test_no_config(self):
        """Test behavior when no configuration is present."""
        with patch("knwl.di.get_config") as mock_get_config:
            mock_get_config.return_value = None

            @inject_config("non.existent.key")
            def test_function(key=None):
                return key

            result = test_function()
            assert result is None
            mock_get_config.assert_called_once_with("non", "existent", "key")

    def test_service_decorator_basic(self):
        """Test basic service injection."""
        mock_service = Mock()

        with patch("knwl.di.services") as mock_services:
            mock_services.create_service.return_value = mock_service

            @service("llm")
            def test_function(text: str, llm=None):
                print("Inside test_function")
                return llm.process(text)

            result = test_function("test text")

            mock_services.create_service.assert_called_once_with(
                "llm", variant_name=None, override=None
            )
            mock_service.process.assert_called_once_with("test text")

    def test_service_decorator_with_variant(self):
        """Test service injection with variant specification."""
        mock_service = Mock()

        with patch("knwl.di.services") as mock_services:
            mock_services.create_service.return_value = mock_service

            @service("llm", variant="ollama", param_name="abc")
            def test_function(text: str, abc=None):
                return abc.generate(text)

            result = test_function("test text")

            mock_services.create_service.assert_called_once_with(
                "llm", variant_name="ollama", override=None
            )
            mock_service.generate.assert_called_once_with("test text")

    def test_singleton_service_decorator(self):
        """Test singleton service injection."""
        mock_service = Mock()

        with patch("knwl.di.services") as mock_services:
            mock_services.get_service.return_value = mock_service

            @singleton_service("graph", variant="nx")
            def test_function(data: dict, graph=None):
                return graph.add_node(data)

            result = test_function({"id": "1", "name": "test"})

            mock_services.get_service.assert_called_once_with(
                "graph", variant_name="nx", override=None
            )
            mock_service.add_node.assert_called_once_with({"id": "1", "name": "test"})

    def test_inject_config_decorator(self):
        """Test configuration value injection."""
        with patch("knwl.di.get_config") as mock_get_config:
            mock_get_config.side_effect = lambda *keys: {
                ("api", "host"): "localhost",
                ("api", "port"): 8080,
            }.get(keys, None)

            @inject_config("api.host", "api.port")
            def test_function(host=None, port=None):
                return f"Server: {host}:{port}"

            result = test_function()

            assert result == "Server: localhost:8080"
            assert mock_get_config.call_count == 2

    def test_inject_services_decorator(self):
        """Test multiple service injection with inject_services."""
        mock_llm = Mock()
        mock_vector = Mock()

        with patch("knwl.di.services") as mock_services:

            def side_effect(service_name, variant_name=None, override=None):
                if service_name == "llm":
                    return mock_llm
                elif service_name == "vector":
                    return mock_vector
                return Mock()

            mock_services.create_service.side_effect = side_effect
            mock_services.get_service.side_effect = side_effect

            @inject_services(
                llm="llm",
                storage=("vector", "chroma"),
                graph={"service": "graph", "variant": "nx", "singleton": True},
            )
            def test_function(data: str, llm=None, storage=None, graph=None):
                return {
                    "llm_result": llm.process(data),
                    "storage_result": storage.store(data),
                    "graph_nodes": graph.number_of_nodes(),
                }

            result = test_function("test data")

            # Verify services were called correctly
            mock_llm.process.assert_called_once_with("test data")
            mock_vector.store.assert_called_once_with("test data")

    def test_auto_inject_decorator(self):
        """Test automatic injection based on parameter names."""
        mock_llm = Mock()
        mock_vector = Mock()

        with patch("knwl.di.services") as mock_services:

            def side_effect(service_name, variant_name=None, override=None):
                if service_name == "llm":
                    return mock_llm
                elif service_name == "vector":
                    return mock_vector
                return Mock()

            mock_services.get_service.side_effect = side_effect

            @auto_inject
            def test_function(text: str, llm=None, vector=None, other_param=None):
                results = {}
                if llm:
                    results["llm"] = llm.analyze(text)
                if vector:
                    results["vector"] = vector.search(text)
                return results

            result = test_function("test text")

            mock_llm.analyze.assert_called_once_with("test text")
            mock_vector.search.assert_called_once_with("test text")

    def test_service_provider_create_service(self):
        """Test ServiceProvider.create_service method."""
        mock_service = Mock()

        with patch("knwl.di.services") as mock_services:
            mock_services.create_service.return_value = mock_service

            service_instance = ServiceProvider.create_service("llm", variant="ollama")

            mock_services.create_service.assert_called_once_with(
                "llm", variant_name="ollama", override=None
            )
            assert service_instance == mock_service

    def test_service_provider_get_service(self):
        """Test ServiceProvider.get_service method."""
        mock_service = Mock()

        with patch("knwl.di.services") as mock_services:
            mock_services.get_service.return_value = mock_service

            service_instance = ServiceProvider.get_service("graph", variant="nx")

            mock_services.get_service.assert_called_once_with(
                "graph", variant_name="nx", override=None
            )
            assert service_instance == mock_service

    def test_service_provider_clear_singletons(self):
        """Test ServiceProvider.clear_singletons method."""
        with patch("knwl.di.services") as mock_services:
            ServiceProvider.clear_singletons()
            mock_services.clear_singletons.assert_called_once()

    def test_di_container_registration(self):
        """Test DI container service registration."""
        container.register_service_injection(
            "test_func", "llm", "ollama", "language_model", singleton=True
        )

        assert "test_func" in container._service_registry
        assert "language_model" in container._service_registry["test_func"]

        service_info = container._service_registry["test_func"]["language_model"]
        assert service_info["service_name"] == "llm"
        assert service_info["variant_name"] == "ollama"
        assert service_info["singleton"] is True

    def test_di_container_config_registration(self):
        """Test DI container config registration - verifies the registration mechanism."""
        container.register_config_injection("test_func", ["api.host", "api.port"])

        assert "test_func" in container._config_registry
        assert container._config_registry["test_func"]["config_keys"] == [
            "api.host",
            "api.port",
        ]
        assert container._config_registry["test_func"]["override"] is None

    def test_manual_config_injection_with_wrapper(self):
        """Test manual config registration with actual injection via wrapper."""
        with patch("knwl.di.get_config") as mock_get_config:
            mock_get_config.side_effect = lambda *keys: {
                ("api", "port"): 9000,
                ("api", "host"): "manual.host.com",
            }.get(keys, None)

            def my_function(host=None, port=None):
                return {"host": host, "port": port}

            # Manually register config injection
            func_name = f"{my_function.__module__}.{my_function.__qualname__}"
            container.register_config_injection(func_name, ["api.host", "api.port"])

            # To actually use manual registration, you need to manually call inject_dependencies
            injected_args = container.inject_dependencies(my_function)
            result = my_function(**injected_args)

            assert result == {"host": "manual.host.com", "port": 9000}
            assert mock_get_config.call_count == 2

    def test_config_injection_with_nested_keys(self):
        """Test config injection with deeply nested configuration keys."""
        with patch("knwl.di.get_config") as mock_get_config:
            mock_get_config.side_effect = lambda *keys: {
                ("database", "connection", "host"): "db.example.com",
                ("database", "connection", "port"): 5432,
                ("logging", "level"): "INFO",
            }.get(keys, None)

            @inject_config(
                "database.connection.host", "database.connection.port", "logging.level"
            )
            def test_function(host=None, port=None, level=None):
                return {"db": f"{host}:{port}", "log_level": level}

            result = test_function()

            assert result == {"db": "db.example.com:5432", "log_level": "INFO"}
            assert mock_get_config.call_count == 3

    def test_config_injection_parameter_name_mapping(self):
        """Test that config keys are correctly mapped to parameter names."""
        container.register_config_injection(
            "test_func", ["api.server.host", "api.server.port"]
        )

        assert "test_func" in container._config_registry
        assert (
            "api.server.host" in container._config_registry["test_func"]["config_keys"]
        )
        assert (
            "api.server.port" in container._config_registry["test_func"]["config_keys"]
        )

    def test_config_injection_with_none_values(self):
        """Test config injection when config values are None."""
        with patch("knwl.di.get_config") as mock_get_config:
            mock_get_config.return_value = None

            @inject_config("missing.config.key")
            def test_function(key=None):
                return key

            result = test_function()
            assert result is None

    def test_config_injection_multiple_registrations(self):
        """Test multiple config registrations for the same function."""
        container.register_config_injection("func1", ["key1", "key2"])
        container.register_config_injection("func1", ["key3", "key4"])

        # Second registration should replace the first
        assert container._config_registry["func1"]["config_keys"] == ["key3", "key4"]

    def test_config_injection_empty_key_list(self):
        """Test config injection with empty key list."""
        container.register_config_injection("test_func", [])

        assert "test_func" in container._config_registry
        assert container._config_registry["test_func"]["config_keys"] == []

    def test_mixed_service_and_config_injection(self):
        """Test function with both service and config injection."""
        mock_service = Mock()

        with (
            patch("knwl.di.services") as mock_services,
            patch("knwl.di.get_config") as mock_get_config,
        ):

            mock_services.create_service.return_value = mock_service
            mock_get_config.side_effect = lambda *keys: {
                ("api", "timeout"): 30,
                ("api", "retries"): 3,
            }.get(keys, None)

            @service("llm")
            @inject_config("api.timeout", "api.retries")
            def test_function(text: str, llm=None, timeout=None, retries=None):
                return {
                    "result": llm.process(text),
                    "timeout": timeout,
                    "retries": retries,
                }

            result = test_function("test text")

            mock_services.create_service.assert_called_once()
            assert mock_get_config.call_count == 2
            assert result["timeout"] == 30
            assert result["retries"] == 3

    def test_config_injection_with_default_values(self):
        """Test that config injection respects function default values."""
        with patch("knwl.di.get_config") as mock_get_config:
            mock_get_config.side_effect = lambda *keys: {("server", "port"): 8080}.get(
                keys, "default_value"
            )

            @inject_config("server.host", "server.port")
            def test_function(host="localhost", port=3000):
                return f"{host}:{port}"

            result = test_function()

            # Function defaults should be preserved (DI doesn't override defaults)
            assert result == "localhost:3000"

            # Config injection should only happen for None defaults
            @inject_config("server.host", "server.port")
            def test_function_with_none_defaults(host=None, port=None):
                return f"{host}:{port}"

            result2 = test_function_with_none_defaults()
            assert result2 == "default_value:8080"

    def test_config_injection_partial_override(self):
        """Test config injection when only some parameters are provided."""
        with patch("knwl.di.get_config") as mock_get_config:
            mock_get_config.side_effect = lambda *keys: {
                ("api", "host"): "configured.host.com"
            }.get(keys, None)

            @inject_config("api.host", "api.port")
            def test_function(host=None, port=None):
                return {"host": host, "port": port}

            # Pass port explicitly, should not be injected
            result = test_function(port=9000)

            assert result["host"] == "configured.host.com"
            assert result["port"] == 9000
            # Only host should be looked up in config
            mock_get_config.assert_called_once_with("api", "host")

    def test_config_injection_function_qualname_handling(self):
        """Test that function qualname is handled correctly for nested functions."""

        def outer_function():
            @inject_config("test.key")
            def inner_function(key=None):
                return key

            return inner_function

        inner_func = outer_function()
        func_name = f"{inner_func.__module__}.{inner_func.__qualname__}"

        # The function should be registered with the correct qualname
        assert func_name in container._config_registry

    def test_config_registry_independence(self):
        """Test that config registry is independent of service registry."""
        container.register_service_injection("func1", "service1")
        container.register_config_injection("func2", ["config.key1"])

        assert "func1" in container._service_registry
        assert "func2" in container._config_registry
        assert "func1" not in container._config_registry
        assert "func2" not in container._service_registry

    def test_parameter_override_prevents_injection(self):
        """Test that explicitly passed parameters prevent injection."""
        mock_service = Mock()
        user_service = Mock()

        with patch("knwl.di.services") as mock_services:
            mock_services.create_service.return_value = mock_service

            @service("llm")
            def test_function(text: str, llm=None):
                return llm.process(text)

            # Pass explicit service - should not inject
            result = test_function("test text", llm=user_service)

            # Should not call services.create_service
            mock_services.create_service.assert_not_called()
            # Should use the user-provided service
            user_service.process.assert_called_once_with("test text")

    def test_error_handling_service_injection(self):
        """Test error handling when service injection fails."""
        with patch("knwl.di.services") as mock_services:
            mock_services.create_service.side_effect = Exception(
                "Service creation failed"
            )

            @service("invalid_service")
            def test_function(text: str, invalid_service=None):
                return invalid_service.process(text)

            with pytest.raises(Exception, match="Service creation failed"):
                test_function("test text")

    def test_error_handling_config_injection(self):
        """Test error handling when config injection fails."""
        with patch("knwl.di.get_config") as mock_get_config:
            mock_get_config.side_effect = Exception("Config lookup failed")

            # this invalid key will show up in the Knwl logging
            @inject_config("unit.tests.invalid.config.key")
            def test_function(key=None):
                return key

            with pytest.raises(Exception, match="Config lookup failed"):
                test_function()

    def test_complex_service_specs_validation(self):
        """Test validation of complex service specifications in inject_services."""

        # Test invalid service spec
        with pytest.raises(ValueError, match="Invalid service specification"):

            @inject_services(invalid_param=123)  # Invalid spec type
            def test_function(invalid_param=None):
                pass

    def test_function_signature_binding(self):
        """Test that function signature binding works correctly with various parameter types."""
        mock_service = Mock()

        with patch("knwl.di.services") as mock_services:
            mock_services.create_service.return_value = mock_service

            @service("llm")
            def test_function(
                required_param: str, optional_param: str = "default", llm=None, **kwargs
            ):
                return {
                    "required": required_param,
                    "optional": optional_param,
                    "llm_available": llm is not None,
                    "extra_kwargs": kwargs,
                }

            result = test_function("required_value", extra_arg="extra")

            assert result["required"] == "required_value"
            assert result["optional"] == "default"
            assert result["llm_available"] is True
            assert result["extra_kwargs"] == {"extra_arg": "extra"}

    def test_inject_config_with_override(self):
        """Test inject_config decorator with override parameter."""

        # Test override functionality
        @inject_config(
            "api.host", "api.port", override={"api": {"host": "override.example.com"}}
        )
        def test_function_with_override(host=None, port=None):
            return {"host": host, "port": port}

        with patch("knwl.di.get_config") as mock_get_config:
            mock_get_config.side_effect = lambda *keys: {
                ("api", "port"): 8080,
                ("api", "host"): "original.example.com",
            }[keys]

            result = test_function_with_override()

            # Should use override for host and original config for port
            assert result["host"] == "override.example.com"
            assert result["port"] == 8080
            # get_config should only be called once for port, not for host due to override
            mock_get_config.assert_called_once_with("api", "port")

    def test_config_injection_registration_with_override(self):
        """Test that config injection with override is properly registered."""
        override_dict = {"api": {"host": "test.override.com", "port": 9999}}
        container.register_config_injection(
            "test_func", ["api.host", "api.port"], override=override_dict
        )

        assert "test_func" in container._config_registry
        config_info = container._config_registry["test_func"]
        assert config_info["config_keys"] == ["api.host", "api.port"]
        assert config_info["override"] == override_dict

    def test_nested_override_functionality(self):
        """Test that nested override dictionaries work correctly."""

        @inject_config(
            "llm.model",
            "llm.temperature",
            "api.port",
            override={
                "llm": {"model": "gpt-4", "temperature": 0.8},
                "api": {"port": 3000},
            },
        )
        def test_nested_overrides(model=None, temperature=None, port=None):
            return {"model": model, "temperature": temperature, "port": port}

        with patch("knwl.di.get_config") as mock_get_config:
            # Mock should not be called for any key since all are overridden
            result = test_nested_overrides()

            # All values should come from overrides
            assert result["model"] == "gpt-4"
            assert result["temperature"] == 0.8
            assert result["port"] == 3000
            # get_config should not be called at all since all values are overridden
            mock_get_config.assert_not_called()

    def test_partial_nested_override(self):
        """Test nested overrides with some values from config and some from override."""

        @inject_config(
            "llm.model",
            "llm.temperature",
            "api.host",
            override={
                "llm": {"model": "override-model"}
            },  # Only override model, not temperature
        )
        def test_partial_override(model=None, temperature=None, host=None):
            return {"model": model, "temperature": temperature, "host": host}

        with patch("knwl.di.get_config") as mock_get_config:
            mock_get_config.side_effect = lambda *keys: {
                ("llm", "temperature"): 0.5,
                ("api", "host"): "config.example.com",
            }[keys]

            result = test_partial_override()

            # model should come from override, others from config
            assert result["model"] == "override-model"
            assert result["temperature"] == 0.5
            assert result["host"] == "config.example.com"
            # get_config should be called twice (for temperature and host)
            assert mock_get_config.call_count == 2

    def test_deep_nested_override(self):
        """Test deeply nested override structures."""

        @inject_config(
            "deep.nested.key1",
            "deep.nested.key2",
            override={"deep": {"nested": {"key1": "deep_value"}}},
        )
        def test_deep_override(key1=None, key2=None):
            return {"key1": key1, "key2": key2}

        with patch("knwl.di.get_config") as mock_get_config:
            mock_get_config.side_effect = lambda *keys: {
                ("deep", "nested", "key2"): "config_value"
            }[keys]

            result = test_deep_override()

            assert result["key1"] == "deep_value"
            assert result["key2"] == "config_value"
            mock_get_config.assert_called_once_with("deep", "nested", "key2")

    def test_defaults_decorator_basic(self):
        """Test basic defaults injection from service configuration."""

        @defaults("something")
        class TestClass:
            def __init__(self, llm=None):
                self.llm = llm

        with (
            patch("knwl.di.get_config") as mock_get_config,
            patch("knwl.di.services") as mock_services,
        ):
            # Mock the config lookups
            mock_get_config.side_effect = lambda *args, **kwargs: {
                ("something", "default"): "basic",
                ("something", "basic"): {
                    "class": "test.TestClass",
                    "llm": "@/llm/openai",
                },
            }.get(args)

            # Mock the service instantiation
            mock_llm = Mock()
            mock_services.get_service.return_value = mock_llm

            instance = TestClass()

            assert instance.llm is mock_llm
            mock_services.get_service.assert_called_once_with(
                "llm", variant_name="openai", override=None
            )

    def test_defaults_decorator_with_variant(self):
        """Test defaults injection with explicit variant."""

        @defaults("llm", variant="ollama")
        class TestClass:
            def __init__(self, model=None, temperature=None):
                self.model = model
                self.temperature = temperature

        with patch("knwl.di.get_config") as mock_get_config:
            mock_get_config.return_value = {
                "class": "test.LLM",
                "model": "qwen2.5:14b",
                "temperature": 0.1,
                "context_window": 32768,
            }

            instance = TestClass()

            assert instance.model == "qwen2.5:14b"
            assert instance.temperature == 0.1

    def test_defaults_decorator_parameter_filtering(self):
        """Test that only matching parameters are injected."""

        @defaults("llm")
        class TestClass:
            def __init__(self, model=None, temperature=None):
                self.model = model
                self.temperature = temperature

        with patch("knwl.di.get_config") as mock_get_config:
            # Config has more parameters than constructor accepts
            mock_get_config.side_effect = lambda *args, **kwargs: {
                ("llm", "default"): "ollama",
                ("llm", "ollama"): {
                    "class": "test.LLM",
                    "model": "qwen2.5:14b",
                    "temperature": 0.1,
                    "context_window": 32768,  # Not in constructor
                    "caching": "json",  # Not in constructor
                },
            }.get(args)

            instance = TestClass()

            # Only model and temperature should be injected
            assert instance.model == "qwen2.5:14b"
            assert instance.temperature == 0.1

    def test_defaults_decorator_override(self):
        """Test that provided arguments override defaults."""

        @defaults("llm")
        class TestClass:
            def __init__(self, model=None, temperature=None):
                self.model = model
                self.temperature = temperature

        with patch("knwl.di.get_config") as mock_get_config:
            mock_get_config.side_effect = lambda *args, **kwargs: {
                ("llm", "default"): "ollama",
                ("llm", "ollama"): {
                    "class": "test.LLM",
                    "model": "qwen2.5:14b",
                    "temperature": 0.1,
                },
            }.get(args)

            # Override the model parameter
            instance = TestClass(model="custom-model")

            assert instance.model == "custom-model"  # Overridden
            assert instance.temperature == 0.1  # Injected

    def test_defaults_decorator_service_reference(self):
        """Test that service references (@/) are properly instantiated."""

        @defaults("graph_extraction")
        class TestClass:
            def __init__(self, llm=None, mode=None):
                self.llm = llm
                self.mode = mode

        with (
            patch("knwl.di.get_config") as mock_get_config,
            patch("knwl.di.services") as mock_services,
        ):

            mock_get_config.side_effect = lambda *args, **kwargs: {
                ("graph_extraction", "default"): "basic",
                ("graph_extraction", "basic"): {
                    "class": "test.GraphExtraction",
                    "llm": "@/llm/ollama",
                    "mode": "full",
                },
            }.get(args)

            mock_llm = Mock()
            mock_services.get_service.return_value = mock_llm

            instance = TestClass()

            assert instance.llm is mock_llm
            assert instance.mode == "full"
            mock_services.get_service.assert_called_once_with(
                "llm", variant_name="ollama", override=None
            )

    def test_defaults_decorator_with_inject_config(self):
        """Test combining @defaults with @inject_config."""

        @defaults("graph_extraction")
        @inject_config("api.host", "api.port")
        class TestClass:
            def __init__(self, llm=None, mode=None, host=None, port=None):
                self.llm = llm
                self.mode = mode
                self.host = host
                self.port = port

        with (
            patch("knwl.di.get_config") as mock_get_config,
            patch("knwl.di.services") as mock_services,
        ):

            def config_side_effect(*args, **kwargs):
                if args == ("graph_extraction", "default"):
                    return "basic"
                elif args == ("graph_extraction", "basic"):
                    return {
                        "class": "test.GraphExtraction",
                        "llm": "@/llm/ollama",
                        "mode": "full",
                    }
                elif args == ("api", "host"):
                    return "localhost"
                elif args == ("api", "port"):
                    return 8080
                return None

            mock_get_config.side_effect = config_side_effect
            mock_llm = Mock()
            mock_services.get_service.return_value = mock_llm

            instance = TestClass()

            assert instance.llm is mock_llm
            assert instance.mode == "full"
            assert instance.host == "localhost"
            assert instance.port == 8080

    def test_defaults_decorator_no_default_variant(self):
        """Test handling when no default variant is specified."""

        @defaults("custom_service")
        class TestClass:
            def __init__(self, param=None):
                self.param = param

        with patch("knwl.di.get_config") as mock_get_config:
            # No default variant
            mock_get_config.return_value = None

            # Should not raise an error, just skip injection
            instance = TestClass()
            assert instance.param is None

    def test_inject_config_class_decorator(self):
        """Test inject_config decorator on a class."""

        with patch("knwl.di.get_config") as mock_get_config:
            mock_get_config.side_effect = lambda *keys: {
                ("api", "host"): "class.example.com",
                ("api", "port"): 9000,
            }.get(keys, None)

            @inject_config("api.host", "api.port")
            class TestService:
                def __init__(self, host=None, port=None):
                    self.host = host
                    self.port = port

            instance = TestService()
            assert instance.host == "class.example.com"
            assert instance.port == 9000

    def test_inject_config_class_with_dict_format(self):
        """Test inject_config decorator on a class with dictionary format."""

        with patch("knwl.di.get_config") as mock_get_config:
            mock_get_config.side_effect = lambda *keys: {
                ("chunking", "tiktoken", "model"): "gpt-4",
                ("chunking", "tiktoken", "size"): 512,
            }.get(keys, None)

            @inject_config(
                {
                    "chunking.tiktoken.model": "model",
                    "chunking.tiktoken.size": "chunk_size",
                }
            )
            class ChunkingService:
                def __init__(self, model=None, chunk_size=None):
                    self.model = model
                    self.chunk_size = chunk_size

            instance = ChunkingService()
            assert instance.model == "gpt-4"
            assert instance.chunk_size == 512

    def test_inject_config_class_with_param_name(self):
        """Test inject_config decorator on a class with custom param_name."""

        with patch("knwl.di.get_config") as mock_get_config:
            mock_get_config.side_effect = lambda *keys: {
                ("api", "host"): "custom.host.com",
            }.get(keys, None)

            @inject_config("api.host", param_name="server_host")
            class TestService:
                def __init__(self, server_host=None):
                    self.host = server_host

            instance = TestService()
            assert instance.host == "custom.host.com"

    def test_inject_config_class_dict_with_additional_keys_error(self):
        """Test that dict format with additional keys raises error."""

        with pytest.raises(
            ValueError, match="Cannot use additional config keys with dictionary format"
        ):

            @inject_config({"api.host": "host"}, "api.port")
            class TestService:
                def __init__(self, host=None, port=None):
                    pass

    def test_inject_config_class_dict_with_param_name_error(self):
        """Test that dict format with param_name raises error."""

        with pytest.raises(
            ValueError, match="Cannot use param_name with dictionary format"
        ):

            @inject_config({"api.host": "host"}, param_name="custom")
            class TestService:
                def __init__(self, host=None):
                    pass

    def test_inject_config_function_dict_with_additional_keys_error(self):
        """Test that function decorator dict format with additional keys raises error."""

        with pytest.raises(
            ValueError, match="Cannot use additional config keys with dictionary format"
        ):

            @inject_config({"api.host": "host"}, "api.port")
            def test_function(host=None, port=None):
                pass

    def test_inject_config_function_dict_with_param_name_error(self):
        """Test that function decorator dict format with param_name raises error."""

        with pytest.raises(
            ValueError, match="Cannot use param_name with dictionary format"
        ):

            @inject_config({"api.host": "host"}, param_name="custom")
            def test_function(host=None):
                pass

    def test_inject_config_function_param_name_with_multiple_keys_error(self):
        """Test that param_name with multiple keys raises error."""

        with pytest.raises(
            ValueError, match="param_name can only be used with a single config key"
        ):

            @inject_config("api.host", "api.port", param_name="custom")
            def test_function(custom=None):
                pass

    def test_defaults_decorator_function(self):
        """Test defaults decorator on a function."""

        with patch("knwl.di.get_config") as mock_get_config:
            mock_get_config.side_effect = lambda *args, **kwargs: {
                ("llm", "default"): "ollama",
                ("llm", "ollama"): {
                    "class": "test.LLM",
                    "model": "qwen2.5:14b",
                    "temperature": 0.1,
                },
            }.get(args)

            @defaults("llm")
            def process_text(text, model=None, temperature=None):
                return {"text": text, "model": model, "temperature": temperature}

            result = process_text("hello")
            assert result["model"] == "qwen2.5:14b"
            assert result["temperature"] == 0.1

    def test_defaults_decorator_no_service_config(self):
        """Test defaults decorator when service config is not found."""

        with patch("knwl.di.get_config") as mock_get_config:
            mock_get_config.side_effect = lambda *args, **kwargs: {
                ("unknown_service", "default"): "variant1",
                ("unknown_service", "variant1"): None,  # No config
            }.get(args)

            @defaults("unknown_service")
            def test_function(param=None):
                return param

            result = test_function()
            assert result is None

    def test_defaults_decorator_empty_string_values(self):
        """Test that empty string values are skipped during defaults injection."""

        with patch("knwl.di.get_config") as mock_get_config:
            mock_get_config.side_effect = lambda *args, **kwargs: {
                ("llm", "default"): "ollama",
                ("llm", "ollama"): {
                    "class": "test.LLM",
                    "model": "",  # Empty string should be skipped
                    "temperature": 0.1,
                },
            }.get(args)

            @defaults("llm")
            def test_function(model=None, temperature=None):
                return {"model": model, "temperature": temperature}

            result = test_function()
            assert result["model"] is None  # Should not be injected
            assert result["temperature"] == 0.1

    def test_defaults_decorator_none_string_values(self):
        """Test that 'None' string values are skipped during defaults injection."""

        with patch("knwl.di.get_config") as mock_get_config:
            mock_get_config.side_effect = lambda *args, **kwargs: {
                ("llm", "default"): "ollama",
                ("llm", "ollama"): {
                    "class": "test.LLM",
                    "model": "none",  # 'none' string should be skipped
                    "temperature": 0.1,
                },
            }.get(args)

            @defaults("llm")
            def test_function(model=None, temperature=None):
                return {"model": model, "temperature": temperature}

            result = test_function()
            assert result["model"] is None  # Should not be injected
            assert result["temperature"] == 0.1

    def test_defaults_decorator_explicit_none_conversion(self):
        """Test that already-bound 'None' string values are converted to None."""

        with patch("knwl.di.get_config") as mock_get_config:
            mock_get_config.side_effect = lambda *args, **kwargs: {
                ("llm", "default"): "ollama",
                ("llm", "ollama"): {
                    "class": "test.LLM",
                    "model": "qwen2.5:14b",
                    "temperature": 0.1,
                },
            }.get(args)

            @defaults("llm")
            def test_function(model="None", temperature=None):
                return {"model": model, "temperature": temperature}

            # Pass "None" explicitly - this tests the bound argument conversion
            result = test_function(model="None")
            assert result["model"] is None  # Should be converted
            assert result["temperature"] == 0.1

    def test_safe_bind_partial_with_var_keyword(self):
        """Test safe_bind_partial with **kwargs parameter."""

        def func_with_kwargs(a, b=None, **kwargs):
            return {"a": a, "b": b, "kwargs": kwargs}

        bound = container.safe_bind_partial(
            func_with_kwargs, 1, b=2, extra1="x", extra2="y"
        )
        assert bound.arguments["a"] == 1
        assert bound.arguments["b"] == 2
        # Extra kwargs are collected in the 'kwargs' key
        assert "kwargs" in bound.arguments
        assert bound.arguments["kwargs"]["extra1"] == "x"
        assert bound.arguments["kwargs"]["extra2"] == "y"

    def test_safe_bind_partial_without_var_keyword(self):
        """Test safe_bind_partial filtering out invalid kwargs."""

        def func_without_kwargs(a, b=None):
            return {"a": a, "b": b}

        # Should filter out extra1 and extra2
        bound = container.safe_bind_partial(
            func_without_kwargs, 1, b=2, extra1="x", extra2="y"
        )
        assert bound.arguments["a"] == 1
        assert bound.arguments["b"] == 2
        assert "extra1" not in bound.arguments
        assert "extra2" not in bound.arguments

    def test_safe_bind_partial_with_self(self):
        """Test safe_bind_partial with class method (self parameter)."""

        class TestClass:
            def method(self, a, b=None):
                return {"a": a, "b": b}

        instance = TestClass()
        # When calling safe_bind_partial on a bound method, 'self' is already bound
        # so we just pass 'a' and other params
        bound = container.safe_bind_partial(instance.method, 1, b=2, extra="ignore")
        # The bound method already has 'self', so 'a' gets value 1
        assert "a" in bound.arguments
        assert bound.arguments["a"] == 1
        assert bound.arguments["b"] == 2
        # 'extra' should be filtered out since method doesn't accept it
        assert "extra" not in bound.arguments

    def test_get_override_value_simple(self):
        """Test _get_override_value with simple key."""
        from knwl.di import _get_override_value

        override = {"api": {"host": "test.com", "port": 8080}}
        assert _get_override_value(override, "api.host") == "test.com"
        assert _get_override_value(override, "api.port") == 8080

    def test_get_override_value_deep_nested(self):
        """Test _get_override_value with deeply nested keys."""
        from knwl.di import _get_override_value

        override = {"level1": {"level2": {"level3": {"key": "value"}}}}
        assert _get_override_value(override, "level1.level2.level3.key") == "value"

    def test_get_override_value_missing_key(self):
        """Test _get_override_value with missing key returns default."""
        from knwl.di import _get_override_value

        override = {"api": {"host": "test.com"}}
        assert _get_override_value(override, "api.missing", "default") == "default"
        assert _get_override_value(override, "missing.key", None) is None

    def test_get_override_value_empty_dict(self):
        """Test _get_override_value with empty override dict."""
        from knwl.di import _get_override_value

        assert _get_override_value({}, "any.key", "default") == "default"
        assert _get_override_value(None, "any.key", "default") == "default"

    def test_service_provider_context_manager(self):
        """Test ServiceProvider as context manager."""

        provider = ServiceProvider(api={"host": "override.com"})

        with provider as p:
            assert p is provider
            assert p.overrides == {"api": {"host": "override.com"}}

    def test_inject_config_list_format(self):
        """Test inject_config with list format."""

        with patch("knwl.di.get_config") as mock_get_config:
            mock_get_config.side_effect = lambda *keys: {
                ("api", "host"): "list.host.com",
                ("api", "port"): 7000,
            }.get(keys, None)

            @inject_config(["api.host", "api.port"])
            def test_function(host=None, port=None):
                return {"host": host, "port": port}

            result = test_function()
            assert result["host"] == "list.host.com"
            assert result["port"] == 7000

    def test_inject_config_tuple_format(self):
        """Test inject_config with tuple format."""

        with patch("knwl.di.get_config") as mock_get_config:
            mock_get_config.side_effect = lambda *keys: {
                ("db", "user"): "admin",
                ("db", "password"): "secret",
            }.get(keys, None)

            @inject_config(("db.user", "db.password"))
            def test_function(user=None, password=None):
                return {"user": user, "password": password}

            result = test_function()
            assert result["user"] == "admin"
            assert result["password"] == "secret"

    def test_defaults_skip_class_parameter(self):
        """Test that 'class' parameter is skipped during defaults injection."""

        with patch("knwl.di.get_config") as mock_get_config:
            mock_get_config.side_effect = lambda *args, **kwargs: {
                ("service", "default"): "variant1",
                ("service", "variant1"): {
                    "class": "some.class.Name",  # Should be skipped
                    "param1": "value1",
                },
            }.get(args)

            @defaults("service")
            def test_function(param1=None):
                return {"param1": param1}

            result = test_function()
            assert result["param1"] == "value1"

    def test_wrapper_kwargs_extraction(self):
        """Test that wrapper properly extracts and merges kwargs."""

        mock_service = Mock()

        with patch("knwl.di.services") as mock_services:
            mock_services.create_service.return_value = mock_service

            @service("llm")
            def test_function(text, llm=None, **extra):
                return {"llm": llm, "extra": extra}

            result = test_function("test", custom_arg="value")

            assert result["llm"] is mock_service
            # The wrapper merges kwargs, so extra gets nested
            assert (
                "custom_arg" in result["extra"]
                or result["extra"].get("extra", {}).get("custom_arg") == "value"
            )

    def test_register_defaults_injection(self):
        """Test register_defaults_injection method."""

        container.register_defaults_injection(
            "test_func", "service_name", "variant_name", {"key": "value"}
        )

        assert "test_func" in container._defaults_registry
        defaults_info = container._defaults_registry["test_func"]
        assert defaults_info["service_name"] == "service_name"
        assert defaults_info["variant_name"] == "variant_name"
        assert defaults_info["override"] == {"key": "value"}

    def test_inject_config_with_dict_mapping(self):
        """Test inject_config function decorator with dict mapping format."""

        with patch("knwl.di.get_config") as mock_get_config:
            mock_get_config.side_effect = lambda *keys: {
                ("chunking", "size"): 1024,
                ("chunking", "overlap"): 128,
            }.get(keys, None)

            @inject_config(
                {"chunking.size": "chunk_size", "chunking.overlap": "overlap_size"}
            )
            def test_function(chunk_size=None, overlap_size=None):
                return {"chunk_size": chunk_size, "overlap_size": overlap_size}

            result = test_function()
            assert result["chunk_size"] == 1024
            assert result["overlap_size"] == 128

    def test_legacy_config_injection_with_param_name(self):
        """Test legacy config injection format with custom param_name."""

        with patch("knwl.di.get_config") as mock_get_config:
            mock_get_config.side_effect = lambda *keys: {
                ("api", "endpoint"): "https://api.example.com",
            }.get(keys, None)

            @inject_config("api.endpoint", param_name="url")
            def test_function(url=None):
                return {"url": url}

            result = test_function()
            assert result["url"] == "https://api.example.com"

    def test_auto_inject_unknown_service(self):
        """Test auto_inject with unknown service parameter (should not be injected)."""

        @auto_inject
        def test_function(text: str, unknown_service=None, other_param=None):
            return {"unknown": unknown_service, "other": other_param}

        result = test_function("test")
        assert result["unknown"] is None
        assert result["other"] is None

    def test_inject_services_invalid_tuple_spec(self):
        """Test inject_services with invalid tuple specification."""

        # A tuple with wrong number of elements
        with pytest.raises(ValueError, match="Invalid service specification"):

            @inject_services(param=("service_only",))  # Tuple with only 1 element
            def test_function(param=None):
                pass

    def test_defaults_service_reference_with_variant(self):
        """Test defaults decorator with service reference that has variant."""

        with (
            patch("knwl.di.get_config") as mock_get_config,
            patch("knwl.di.services") as mock_services,
        ):

            mock_get_config.side_effect = lambda *args, **kwargs: {
                ("service", "default"): "variant1",
                ("service", "variant1"): {
                    "class": "test.Service",
                    "llm": "@/llm",  # Service reference without variant
                },
            }.get(args)

            mock_llm = Mock()
            mock_services.get_service.return_value = mock_llm

            @defaults("service")
            def test_function(llm=None):
                return llm

            result = test_function()
            assert result is mock_llm
            # Should be called with variant_name=None when not specified
            mock_services.get_service.assert_called_once_with(
                "llm", variant_name=None, override=None
            )

    def test_inject_services_dict_spec_without_service(self):
        """Test inject_services with dict spec missing 'service' key."""

        @inject_services(param={"variant": "test", "singleton": True})
        def test_function(param=None):
            return param

        # This should handle the case where service_name is None
        # The decorator should still be applied, but injection might fail
        func_name = f"{test_function.__module__}.{test_function.__qualname__}"
        assert func_name in container._service_registry

    def test_config_injection_with_empty_override(self):
        """Test config injection when override dict is empty."""

        with patch("knwl.di.get_config") as mock_get_config:
            mock_get_config.side_effect = lambda *keys: {
                ("api", "host"): "config.host.com",
            }.get(keys, None)

            @inject_config("api.host", override={})
            def test_function(host=None):
                return host

            result = test_function()
            assert result == "config.host.com"

    def test_defaults_with_none_value(self):
        """Test defaults decorator when config value is None."""

        with patch("knwl.di.get_config") as mock_get_config:
            mock_get_config.side_effect = lambda *args, **kwargs: {
                ("service", "default"): "variant1",
                ("service", "variant1"): {
                    "class": "test.Service",
                    "param1": None,  # Explicitly None
                    "param2": "value2",
                },
            }.get(args)

            @defaults("service")
            def test_function(param1=None, param2=None):
                return {"param1": param1, "param2": param2}

            result = test_function()
            # param1 should remain None (not injected)
            assert result["param1"] is None
            assert result["param2"] == "value2"

    def test_service_with_override(self):
        """Test service decorator with override parameter."""

        mock_service = Mock()

        with patch("knwl.di.services") as mock_services:
            mock_services.create_service.return_value = mock_service

            override_config = {"llm": {"model": "custom-model"}}

            @service("llm", override=override_config)
            def test_function(llm=None):
                return llm

            result = test_function()

            mock_services.create_service.assert_called_once_with(
                "llm", variant_name=None, override=override_config
            )
            assert result is mock_service

    def test_singleton_service_with_override(self):
        """Test singleton_service decorator with override parameter."""

        mock_service = Mock()

        with patch("knwl.di.services") as mock_services:
            mock_services.get_service.return_value = mock_service

            override_config = {"graph": {"backend": "networkx"}}

            @singleton_service("graph", override=override_config)
            def test_function(graph=None):
                return graph

            result = test_function()

            mock_services.get_service.assert_called_once_with(
                "graph", variant_name=None, override=override_config
            )
            assert result is mock_service

    def test_inject_services_with_override(self):
        """Test inject_services with override in dict spec."""

        mock_llm = Mock()

        with patch("knwl.di.services") as mock_services:
            mock_services.create_service.return_value = mock_llm

            @inject_services(
                llm={
                    "service": "llm",
                    "variant": "openai",
                    "singleton": False,
                    "override": {"llm": {"model": "gpt-4"}},
                }
            )
            def test_function(llm=None):
                return llm

            result = test_function()
            assert result is mock_llm

    def test_defaults_class_with_override(self):
        """Test defaults decorator on class with override."""

        with patch("knwl.di.get_config") as mock_get_config:
            mock_get_config.side_effect = lambda *args, **kwargs: {
                ("llm", "default"): "ollama",
                ("llm", "ollama"): {
                    "class": "test.LLM",
                    "model": "qwen2.5:14b",
                },
            }.get(args)

            override_config = {"llm": {"ollama": {"model": "override-model"}}}

            @defaults("llm", override=override_config)
            class TestClass:
                def __init__(self, model=None):
                    self.model = model

            instance = TestClass()
            # Should use config value since override doesn't match the structure
            assert instance.model == "qwen2.5:14b"

    def test_register_config_injection_dict_format(self):
        """Test register_config_injection with dict mapping format."""

        config_mapping = {"api.host": "server_host", "api.port": "server_port"}

        container.register_config_injection(
            "test_func",
            config_mapping,
            param_name=None,
            override={"api": {"host": "override.com"}},
        )

        assert "test_func" in container._config_registry
        config_info = container._config_registry["test_func"]
        assert config_info["config_mapping"] == config_mapping
        assert config_info["override"] == {"api": {"host": "override.com"}}


if __name__ == "__main__":
    pytest.main([__file__, "-v"])



import os
from pathlib import Path
from abc import ABC, abstractmethod

from knwl.framework_base import FrameworkBase
from knwl.models.KnwlChunk import KnwlChunk


class SummarizationBase(FrameworkBase):

    @abstractmethod
    async def summarize(
        self, content: str | list[str], entity_or_relation_name: str|list[str] = None
    ) -> str:
        """
        Summarize the given content.
        Args:
            content: str | list[str]: The content to summarize.
            entity_or_relation_name:  str | list[str]: The name of the entity or relation to use for the summary.

        Returns:
            str: The summarized content.

        """
        pass



from typing import Optional
from pydantic import BaseModel, Field


class KnwlText(BaseModel):
    """
    Represents a piece of text obtained via one of the Grag strategies.

    Attributes:
        origin_id (str): The source id of the text.
        text (str): The actual text content.
        order (int): The order of the text in the sequence.
        id (str): The unique identifier of the text.
    """

    origin_id: Optional[str] = Field(
        default=None, description="The source id of the text."
    )
    text: Optional[str] = Field(
        default="",
        description="The text content to use for context. If not supplied, downstream services need to fetch the text data separately using the id.",
    )
    index: Optional[int] = Field(
        default=0, description="The order of the text in the sequence."
    )
    id: Optional[str] = Field(
        default=None, description="The unique identifier of the text."
    )

    @staticmethod
    def get_header():
        return ["id", "content"]

    def to_row(self):
        return "\t".join([self.origin_id or "", self.text])
 


from knwl.models import KnwlInput
from knwl.models.KnwlEdge import KnwlEdge
from knwl.models.KnwlReference import KnwlReference

from pydantic import BaseModel, Field
from typing import List

from knwl.models.KnwlText import KnwlText
from knwl.models.KnwlNode import KnwlNode


class KnwlContext(BaseModel):
    """
    Represents the augmented context based on the knowledge graph for a given input.
    """

    input: str | KnwlInput = Field(
        description="The original input text or KnwlInput object."
    )
    texts: list[KnwlText] = Field(default_factory=list)
    nodes: list[KnwlNode] = Field(default_factory=list)
    edges: list[KnwlEdge] = Field(default_factory=list)
    references: list[KnwlReference] = Field(default_factory=list)

    @staticmethod
    def combine(
        first: "KnwlContext", second: "KnwlContext"
    ) -> "KnwlContext":
        texts = [c for c in first.texts]
        nodes = [n for n in first.nodes]
        edges = [e for e in first.edges]
        references = [r for r in first.references]
        # ================= Texts ===========================================
        chunk_ids = [cc.id for cc in texts]
        for c in second.texts:
            if c.id not in chunk_ids:
                texts.append(c)
        # ================= Nodes  ===========================================
        node_ids = [cc.id for cc in nodes]
        for n in second.nodes:
            if n.id not in node_ids:
                nodes.append(n)
        # ================= Edges  ===========================================
        edge_ids = [cc.id for cc in edges]
        for e in second.edges:
            if e.id not in edge_ids:
                edges.append(e)

        # ================= References  ======================================
        reference_ids = [cc.id for cc in references]
        for r in second.references:
            if r.id not in reference_ids:
                references.append(r)

        return KnwlContext(
            input=first.input,
            texts=texts,
            nodes=nodes,
            edges=edges,
            references=references,
        )

    @staticmethod
    def empty(input: KnwlInput) -> "KnwlContext":
        return KnwlContext(
            input=input,
            texts=[],
            nodes=[],
            edges=[],
            references=[],
        )

    def __repr__(self):
        return f"<KnwlContext, texts={len(self.texts)}, nodes={len(self.nodes)}, edges={len(self.edges)}, references={len(self.references)}>"
    def __str__(self):
        return self.__repr__()


"""
Rich terminal formatter for Knwl models.

This module provides a beautiful terminal output using the Rich library,
with consistent styling and formatting across all Knwl models.
"""

import json
from typing import Any, Optional, List, Dict
from pydantic import BaseModel

from rich.console import Console
from rich.panel import Panel
from rich.table import Table
from rich.tree import Tree
from rich.text import Text
from rich.syntax import Syntax
from rich.markdown import Markdown
from rich import box

from knwl.format.formatter_base import FormatterBase, ModelFormatter, get_registry


class RichTheme:
    """Centralized theme configuration for consistent styling."""

    # Colors
    PRIMARY = "cyan"
    SECONDARY = "blue"
    SUCCESS = "green"
    WARNING = "yellow"
    ERROR = "red"
    MUTED = "dim"
    HIGHLIGHT = "magenta"

    # Styles
    TITLE_STYLE = "bold cyan"
    SUBTITLE_STYLE = "bold blue"
    KEY_STYLE = "bold yellow"
    VALUE_STYLE = "white"
    TYPE_STYLE = "italic green"
    ID_STYLE = "dim cyan"

    # Box styles
    PANEL_BOX = box.ROUNDED
    TABLE_BOX = box.SIMPLE

    # Borders
    BORDER_STYLE = "cyan"


class RichFormatter(FormatterBase):
    """
    Formatter for rich terminal output.

    This formatter uses the Rich library to create beautiful,
    colorful terminal output with tables, panels, and trees.
    """

    def __init__(self, console: Optional[Console] = None):
        """
        Initialize the Rich formatter.

        Args:
            console: Optional Rich Console instance. If not provided, creates a new one.
        """
        self.console = console or Console(width=80)
        self.theme = RichTheme()
        self._registry = get_registry()

    def format(self, obj: Any, **options) -> Any:
        """
        Format an object using Rich components.

        Args:
            obj: The object to format
            **options: Formatting options

        Returns:
            Rich renderable object (Panel, Table, Tree, etc.)
        """
        # Handle None
        if obj is None:
            return Text("None", style=self.theme.MUTED)
        if isinstance(obj, str):
            if obj.strip().startswith("@/"):
                from knwl.config import resolve_config

                return json.dumps(resolve_config(obj.strip()), indent=2)
            elif obj.strip().startswith("$/"):
                from knwl.utils import get_full_path

                path = get_full_path(obj.strip())
                return path
            else:
                return f"To render a configuration, use the '@/service/variant' syntax. For example, `@/vector/memory', '@/llm/openai' or simply '@/' for the whole config."
        # Handle Pydantic models
        if isinstance(obj, BaseModel):
            formatter_class = self._registry.get_formatter(type(obj), "terminal")
            if formatter_class:
                formatter = formatter_class()
                return formatter.format(obj, self, **options)
            else:
                # Fall back to default Pydantic formatting
                return self._format_default_model(obj, **options)

        # Handle lists
        if isinstance(obj, list):
            return self._format_list(obj, **options)

        # Handle dicts
        if isinstance(obj, dict):
            return self._format_dict(obj, **options)

        # Handle primitives
        return self._format_primitive(obj)

    def render(self, obj: Any, **options) -> None:
        """
        Render an object to the console.

        Args:
            obj: The object to render
            **options: Rendering options
        """
        formatted = self.format(obj, **options)
        self.console.print(formatted)

    def _format_default_model(self, model: BaseModel, **options) -> Panel:
        """Format a Pydantic model with default styling."""
        title = options.get("title", model.__class__.__name__)

        table = Table(
            show_header=False, box=self.theme.TABLE_BOX, padding=(0, 1), expand=True
        )
        table.add_column("Field", style=self.theme.KEY_STYLE)
        table.add_column("Value", style=self.theme.VALUE_STYLE)

        for field_name, field_value in model.model_dump().items():
            value_str = self._format_value_for_table(field_value)
            table.add_row(field_name, value_str)

        return Panel(
            table,
            title=f"[{self.theme.TITLE_STYLE}]{title}[/{self.theme.TITLE_STYLE}]",
            border_style=self.theme.BORDER_STYLE,
            box=self.theme.PANEL_BOX,
        )

    def _format_list(self, items: List, **options) -> Any:
        """Format a list of items."""
        if not items:
            return Text("[]", style=self.theme.MUTED)

        # If all items are models, create a compact representation
        if all(isinstance(item, BaseModel) for item in items):
            return self._format_model_list(items, **options)

        # Otherwise, format as bullet list
        lines = []
        for item in items:
            lines.append(f"• {item}")
        return "\n".join(lines)

    def _format_model_list(self, models: list[BaseModel], **options) -> Table:
        """Format a list of Pydantic models as a table."""
        if not models:
            return Text("No items", style=self.theme.MUTED)

        model_type = type(models[0])
        fields = list(models[0].model_fields.keys())

        # Limit fields if too many
        max_fields = options.get("max_fields", 5)
        if len(fields) > max_fields:
            fields = fields[:max_fields]

        table = Table(
            box=self.theme.TABLE_BOX,
            show_header=True,
            header_style=self.theme.SUBTITLE_STYLE,
        )

        for field in fields:
            table.add_column(field.replace("_", " ").title())

        for model in models:
            values = []
            for field in fields:
                value = getattr(model, field, None)
                values.append(self._format_value_for_table(value))
            table.add_row(*values)

        return table

    def _format_dict(self, data: Dict, **options) -> Table:
        """Format a dictionary as a table."""
        table = Table(show_header=False, box=self.theme.TABLE_BOX, padding=(0, 1))
        table.add_column("Key", style=self.theme.KEY_STYLE)
        table.add_column("Value", style=self.theme.VALUE_STYLE)

        for key, value in data.items():
            value_str = self._format_value_for_table(value)
            table.add_row(str(key), value_str)

        return table

    def _format_primitive(self, value: Any) -> Text:
        """Format primitive values."""
        if isinstance(value, bool):
            return Text(
                str(value), style=self.theme.SUCCESS if value else self.theme.ERROR
            )
        elif isinstance(value, (int, float)):
            return Text(str(value), style=self.theme.HIGHLIGHT)
        elif isinstance(value, str):
            return Text(value, style=self.theme.VALUE_STYLE)
        else:
            return Text(str(value), style=self.theme.MUTED)

    def _format_value_for_table(self, value: Any, max_length: int = 150) -> str:
        """Format a value for display in a table cell."""
        if value is None:
            return "[dim]None[/dim]"
        elif isinstance(value, bool):
            style = self.theme.SUCCESS if value else self.theme.ERROR
            return f"[{style}]{value}[/{style}]"
        elif isinstance(value, list):
            if len(value) == 0:
                return "[dim][][/dim]"
            elif len(value) <= 3:
                return f"[{self.theme.MUTED}][{len(value)} items][/{self.theme.MUTED}]"
            else:
                return f"[{self.theme.MUTED}][{len(value)} items][/{self.theme.MUTED}]"
        elif isinstance(value, dict):
            return f"[{self.theme.MUTED}]{{{len(value)} keys}}[/{self.theme.MUTED}]"
        elif isinstance(value, BaseModel):
            return f"[{self.theme.TYPE_STYLE}]{value.__class__.__name__}[/{self.theme.TYPE_STYLE}]"
        else:
            str_value = str(value)
            if len(str_value) > max_length:
                str_value = str_value[: max_length - 3] + "..."
            return str_value

    def create_panel(
        self, content: Any, title: str, subtitle: Optional[str] = None, **kwargs
    ) -> Panel:
        """
        Create a styled panel.

        Args:
            content: The content to display in the panel
            title: Panel title
            subtitle: Optional subtitle
            **kwargs: Additional panel options
        """
        return Panel(
            content,
            title=f"[{self.theme.TITLE_STYLE}]{title}[/{self.theme.TITLE_STYLE}]",
            subtitle=(
                f"[{self.theme.MUTED}]{subtitle}[/{self.theme.MUTED}]"
                if subtitle
                else None
            ),
            border_style=kwargs.get("border_style", self.theme.BORDER_STYLE),
            box=kwargs.get("box", self.theme.PANEL_BOX),
            **{k: v for k, v in kwargs.items() if k not in ["border_style", "box"]},
        )

    def create_table(
        self, title: Optional[str] = None, columns: Optional[list[str]] = None, **kwargs
    ) -> Table:
        """
        Create a styled table.

        Args:
            title: Optional table title
            columns: List of column names
            **kwargs: Additional table options
        """
        table = Table(
            title=(
                f"[{self.theme.SUBTITLE_STYLE}]{title}[/{self.theme.SUBTITLE_STYLE}]"
                if title
                else None
            ),
            box=kwargs.get("box", self.theme.TABLE_BOX),
            show_header=kwargs.get("show_header", True),
            header_style=kwargs.get("header_style", self.theme.SUBTITLE_STYLE),
            **{
                k: v
                for k, v in kwargs.items()
                if k not in ["box", "show_header", "header_style"]
            },
        )

        if columns:
            for col in columns:
                table.add_column(col)

        return table

    def create_tree(self, label: str, **kwargs) -> Tree:
        """
        Create a styled tree.

        Args:
            label: Root label
            **kwargs: Additional tree options
        """
        return Tree(
            f"[{self.theme.TITLE_STYLE}]{label}[/{self.theme.TITLE_STYLE}]", **kwargs
        )



from knwl.di import defaults
from knwl.extraction.keywords_extraction_base import KeywordsExtractionBase
from knwl.llm.llm_base import LLMBase
from knwl.models import KnwlKeywords
from knwl.prompts import prompts
from knwl.utils import hash_with_prefix
import json
from knwl.logging import log


@defaults("keywords_extraction")
class BasicKeywordsExtraction(KeywordsExtractionBase):
    """
    Basic implementation of KeywordsExtractionBase that extracts keywords using simple splitting.

    Methods:
        extract: Extracts keywords from text by splitting on whitespace and punctuation.

    Note:
        This is a basic implementation and may not handle complex keyword extraction scenarios.
    """

    def __init__(self, llm: LLMBase=None):
        super().__init__()
        self.llm = llm
        if not self.llm:
            raise ValueError(
                "BasicKeywordsExtraction: LLM instance is required for extraction."
            )
        if not isinstance(self.llm, LLMBase):
            raise TypeError(
                "BasicKeywordsExtraction: llm must be an instance of LLMBase."
            )

    async def extract(self, text: str, chunk_id: str = None) -> KnwlKeywords | None:
        prompt = prompts.extraction.keywords_extraction(text=text)
        result = await self.llm.ask(prompt)

        try:
            keywords_data = json.loads(result.answer)
            low_keywords = keywords_data.get("low_level_keywords", [])
            high_keywords = keywords_data.get("high_level_keywords", [])
            return KnwlKeywords(
                low_level=low_keywords,
                high_level=high_keywords,
            )
        except json.JSONDecodeError:

            log.warning(
                "BasicKeywordsExtraction: Failed to parse keywords extraction result as JSON."
            )
            return None



from torch import StorageBase
from knwl.di import defaults
from knwl.models.KnwlDocument import KnwlDocument
from knwl.semantic.rag.document_base import DocumentBase
from knwl.storage.storage_adapter import StorageAdapter


@defaults("document_store")
class DocumentStore(DocumentBase):
    """
    Default implementation of DocumentBase.
    This only CRUDs documents to/from storage, if you want auto-chunking and embeddings,
    use the `RagStore` instead.
    """

    def __init__(
        self,
        document_storage: StorageBase | None = None,
    ):
        super().__init__()
        self.document_storage: StorageBase = document_storage
        if self.document_storage is None:
            raise ValueError(f"DocumentStore: document_storage is required.")

    async def upsert(self, obj: str | KnwlDocument) -> str:
        if isinstance(obj, str):
            document = KnwlDocument(content=str(obj).strip())
        else:
            document = obj

        await StorageAdapter.upsert(document, self.document_storage)
        return document.id

    async def get_by_id(self, document_id: str) -> KnwlDocument | None:
        found = await self.document_storage.get_by_id(document_id)
        if found is None:
            return None
        elif isinstance(found, KnwlDocument):
            return found
        else:
            return KnwlDocument.model_validate(found)

    async def delete_by_id(self, document_id: str) -> None:
        await self.document_storage.delete_by_id(document_id)

    async def exists(self, document_id: str) -> bool:
        return await self.document_storage.exists(document_id)



from abc import ABC, abstractmethod
from typing import Union

from pydantic import BaseModel

from knwl.framework_base import FrameworkBase
from knwl.storage.storage_base import StorageBase


class GraphStorageBase(StorageBase, ABC):
    """
    Base class for graph storage implementations.
    Strictly speaking, the graph is a directed multigraph, meaning multiple edges can exist between the same pair of nodes.
    This abstract base class does not use the KnwlNode or KnwlEdge models directly, but the data is expected to be in the form of dictionaries
    that can be converted to/from those models. That is, you can use a GraphStorageBase independently of the higher level mechanisms in knwl.

    - The 'id' (string) should be unique for nodes and edges and present in the data (payload dictionary). If not provided, a uuid4 will be generated.
    - The 'name' (optional string) is not unique and can be the same for different nodes (e.g. "Apple" as fruit and as company).
    - The 'type' (string) is a semantic type (e.g. Person, Company, Location, Product, Event, etc.) and is optional. If not provided, it will be set to "Unknown".

    """

    @abstractmethod
    async def node_exists(self, node_id: str) -> bool:
        """
        Check if a node with the given Id exists in the graph.

        Args:
            node_id (str): The unique identifier of the node to check for existence.

        Returns:
            bool: True if the node exists in the graph, False otherwise.
        """
        ...

    @abstractmethod
    async def edge_exists(
        self,
        source_or_key: Union[BaseModel, str, dict],
        target_node_id: Union[BaseModel, str, dict] = None,
    ) -> bool:
        """
        Check if an edge exists between the given source and target nodes.

        Args:
            source_or_key ( Union[BaseModel,str,dict]): The unique identifier of the source node or edge.
            target_node_id ( Union[BaseModel,str,dict]): The unique identifier of the target node. If not provided, source_or_key is treated as the edge Id.

        Returns:
            bool: True if the edge exists, False otherwise.
        """
        ...

    @abstractmethod
    async def get_node_by_id(self, node_id: str) -> Union[dict, None]:
        """
        Retrieve a node from the graph by its unique identifier.

        Args:
            node_id (str): The unique identifier of the node to retrieve.

        Returns:
            Union[dict, None]: A dictionary containing the node data if found,
                              None if the node does not exist.
        """
        ...

    @abstractmethod
    async def get_nodes_by_name(self, node_name) -> Union[list[dict], None]:
        """
        Retrieves node(s) by its name.

        Args:
            node_name (str): The name of the node to retrieve.

        Returns:
            list[dict] | None: Since the name is not unique and can appear with different semantic types (e.g. Apple as fruit and as company), a list of dictionaries is returned if found, None otherwise.

        """
        ...

    @abstractmethod
    async def node_degree(self, node_id: str) -> int:
        """
        Retrieve the degree of a node in the graph.

        Args:
            node_id (str): The unique identifier of the node.

        Returns:
            int: The degree of the node, or 0 if the node does not exist.
        """
        ...

    @abstractmethod
    async def edge_degree(self, edge_or_source_id: str, target_id: str) -> int:
        """
        Retrieve the degree of an edge in the graph. The degree of an edge is defined as the number of connections it has to other edges (via its endpoints).

        Args:
            edge_or_source_id (str): The unique identifier of the source node.
            target_id (str): The unique identifier of the target node.

        Returns:
            int: The degree of the edge, or 0 if the edge does not exist.
        """
        ...

    @abstractmethod
    async def get_edges(
        self, source_node_id_or_key: str, target_node_id: str = None, label: str = None
    ) -> Union[list[dict], None]:
        """
        Retrieve edges from the graph based on the endpoints and optional label.

        Args:
            source_node_id_or_key (str): The unique id of the edge or the Id of the source node for the edge.
            target_node_id (str, optional): The ID of the target node for the edge.
                If None, returns all edges from the source node. Defaults to None.
            label (str, optional): The label/type of the edge to filter by.
                If None, returns edges regardless of label. Defaults to None.

        Returns:
            Union[list[dict], None]: A list of dictionaries representing the edge(s)
                if found, where each dictionary contains edge properties and metadata.
                Returns None if no edges are found matching the criteria.

        Raises:
            NotImplementedError: This is an abstract method that must be implemented
                by subclasses.
        """
        ...

    @abstractmethod
    async def get_node_edges(self, source_node_id: str) -> list[dict] | None:
        """
        Retrieves all edges connected to the given node.

        Args:
            source_node_id (str): The ID of the source node.

        Returns:
            list[dict] | None: A list of edge objects if the node exists, None otherwise.
        """
        ...

    @abstractmethod
    async def get_edges_between_nodes(
        self, source_id: str, target_id: str
    ) -> list[dict]:
        """
        Retrieve edges between two nodes by their IDs from the knowledge graph.
        """
        ...

    @abstractmethod
    async def get_attached_edges(self, nodes):
        """
        Asynchronously retrieves the edges attached to the given nodes.

        Args:
            nodes (list[KnwlNode]): A list of KnwlNode objects for which to retrieve attached edges.

        Returns:
            list[KnwlEdge]: A list of KnwlEdge objects attached to the given nodes.
        """
        ...

    @abstractmethod
    async def get_edge_degrees(self, edges):
        """
        Asynchronously retrieves the degrees of the given edges.

        Args:
            edges (list[KnwlEdge]): A list of KnwlEdge objects for which to retrieve degrees.

        Returns:
            list[int]: A list of degrees for the given edges.
        """
        ...

    @abstractmethod
    async def get_semantic_endpoints(
        self, edge_ids: list[str]
    ) -> dict[str, tuple[str, str]]:
        """
        Given a list of edge Id's, the name of the source and target nodes is returned as tuple (source_name, target_name).
        The keys of the returned dictionary are the edge Id's.

        example return value:
        `
        {
            "edge_id_1": ("source_node_name_1", "target_node_name_1"),
            "edge_id_2": ("source_node_name_2", "target_node_name_2"),
            ...
        }
        `

        Args:
            edge_ids (list[str]): A list of node IDs for which to retrieve names.
        """
        ...

    @abstractmethod
    async def get_edge_by_id(self, edge_id: str) -> Union[dict, None]:
        """
        Retrieve an edge from the graph by its unique identifier.

        Args:
            edge_id (str): The unique identifier of the edge to retrieve.

        Returns:
            Union[dict, None]: A dictionary containing the edge data if found,
                              None if the edge does not exist.
        """
        ...

    @abstractmethod
    async def upsert_node(self, node_id: BaseModel | str | dict, node_data=None):
        """
        Insert or update a node in the graph storage.

        This method creates a new node if it doesn't exist, or updates an existing node
        with the provided data. The operation is performed asynchronously.

        Args:
            node_id (BaseModel|str|dict): Unique identifier for the node to upsert or a dictionary containing node properties.
            node_data (optional): Data to associate with the node. Can be any type
                depending on the storage implementation. Defaults to None.

        Returns:
            None: This method performs the upsert operation but doesn't return a value.
        """
        ...

    @abstractmethod
    async def upsert_edge(self, source_node_id, target_node_id, edge_data=None):
        """
        Upsert (insert or update) an edge in the graph.

        Args:
            source_node_id: The identifier of the source node for the edge
            target_node_id: The identifier of the target node for the edge
            edge_id (str, optional): Unique identifier for the edge. If None, may be auto-generated
            edge_data (optional): Additional data/properties to associate with the edge

        Returns:
            The result of the upsert operation (implementation-specific)
        """
        ...

    @abstractmethod
    async def clear(self):
        """
        Clear all data from the graph storage.

        This method removes all nodes, edges, and associated data from the graph storage,
        effectively resetting it to an empty state.

        Returns:
            None
        """
        ...

    @abstractmethod
    async def node_count(self):
        """
        Get the total number of nodes in the graph.

        Returns:
            int: The total count of nodes in the graph storage.

        Raises:
            NotImplementedError: This method must be implemented by subclasses.
        """
        ...

    @abstractmethod
    async def edge_count(self):
        """
        Get the total number of edges in the graph.

        Returns:
            int: The total count of edges in the graph storage.

        Raises:
            NotImplementedError: This method must be implemented by subclasses.
        """
        ...

    @abstractmethod
    async def remove_node(self, node_id: str):
        """
        Remove a node from the graph storage.

        Args:
            node_id (str): The unique identifier of the node to be removed.

        Raises:
            NotImplementedError: This method must be implemented by subclasses.
            ValueError: If the node_id is invalid or empty.
            KeyError: If the node with the given node_id does not exist.

        Note:
            This method should also handle cleanup of any edges connected to the node
            being removed to maintain graph integrity.
        """
        ...

    @abstractmethod
    async def remove_edge(
        self, source_node_id_or_key: str, target_node_id: str = None, type: str = None
    ):
        """Remove an edge from the graph.

        Args:
            source_node_id_or_key (str): The ID or key of the source node, or an edge key
                                        if target_node_id is None.
            target_node_id (str, optional): The ID of the target node. If None,
                                           source_node_id_or_key is treated as an edge key.

        Raises:
            KeyError: If the specified edge does not exist in the graph.
            ValueError: If the provided node IDs are invalid.

        Returns:
            None

        Note:
            This method removes the edge from the graph storage. If 'source_node_id_or_key'
            is provided alone, it should be a unique edge identifier. If both parameters
            are provided, they represent the source and target nodes of the edge to remove.
        """
        ...

    @abstractmethod
    async def get_edge_weights(
        self, source_node_id_or_key: str, target_node_id: str = None, type: str = None
    ) -> dict[str, float]:
        """
        Get the weights of edges between two nodes in the graph.

        Args:
            source_node_id_or_key (str): The ID or key of the source node.
            target_node_id (str, optional): The ID of the target node. If None,
                the source_node_id_or_key is treated as an edge key. Defaults to None.
            type (str, optional): The type of the edge weight. Defaults to None.
        Returns:
            dict[str, float]: A dictionary where keys are edge types and values are their corresponding weights.
                If no edges are found, an empty dictionary is returned.
        """
        ...

    @abstractmethod
    async def merge(self, nodes: list[dict], edges: list[dict]) -> None:
        """
        Merge another graph into the current graph storage.

        Args:
            nodes (list[dict]): A list of node dictionaries to merge into the graph.
            edges (list[dict]): A list of edge dictionaries to merge into the graph.

        Returns:
            None

        Note:
            This method should handle merging nodes and edges from the provided graph
            into the existing graph storage, ensuring no duplicates and maintaining integrity.
        """
        ...




import os

import pytest

from knwl.config import get_config
from knwl.llm.ollama import OllamaClient
from knwl.services import services
from knwl.utils import get_full_path

pytestmark = pytest.mark.llm


def test_get_spec():

    override_config = {
        "food": {
            "default": "pizza",
            "pizza": {"size": "large", "delivery": "transport/car"},
            "burger": {"size": "medium"},
        },
        "transport": {"car": {"wheels": 4}, "bike": {"wheels": 2}},
    }
    specs = services.get_service_specs("food", override=override_config)
    assert specs["service_name"] == "food"
    assert specs["variant_name"] == "pizza"
    assert specs["size"] == "large"
    with pytest.raises(ValueError):
        services.get_service_specs("unknown_service", override=override_config)
    with pytest.raises(ValueError):
        services.get_service_specs("food", "unknown_variant", override=override_config)

    specs = services.get_service_specs("food/burger", override=override_config)
    assert specs["service_name"] == "food"
    assert specs["variant_name"] == "burger"
    assert specs["size"] == "medium"

    specs = services.get_service_specs("food", "default", override=override_config)
    assert specs["service_name"] == "food"
    assert specs["variant_name"] == "pizza"
    specs = services.get_service_specs("food/default", None, override=override_config)
    assert specs["service_name"] == "food"
    assert specs["variant_name"] == "pizza"
    with pytest.raises(ValueError):
        services.get_service_specs(None, override=override_config)
    with pytest.raises(ValueError):
        # can't have both "/" and variant in one call
        services.get_service_specs("a/b", "c", override=override_config)

    


@pytest.mark.asyncio
async def test_service_simplicity():

    # simply getting the service with default config
    json_service_default = services.get_service("json")
    assert json_service_default.path == get_full_path(
        get_config("json", "basic", "path")
    )
    # it's a singleton
    json_service_default2 = services.get_service("json")
    assert json_service_default.id == json_service_default2.id

    # you can override the config to define a custom path
    json_service_custom = services.get_service(
        "json/custom",
        override={
            "json": {
                "custom": {
                    "class": "knwl.storage.json_storage.JsonStorage",
                    "path": "$/tests/custom.json",
                }
            }
        },
    )
    await json_service_custom.upsert({"key1": {"data": 123}})
    assert os.path.exists(get_full_path("$/tests/custom.json"))


def test_service_instantiation():
    assert services.get_service_specs("llm")
    assert services.get_default_variant_name("llm") == "ollama"
    llm_service = services.instantiate_service("llm")
    assert llm_service is not None
    assert llm_service.__class__.__name__ == "OllamaClient"
    assert llm_service.model == get_config("llm", "ollama", "model")
    assert llm_service.temperature == 0.1
    assert llm_service.context_window == 32768

    # Test with override configuration
    override_config = {
        "llm": {
            "default": "ollama",
            "ollama": {
                "model": "custom_model:1b",
                "temperature": 0.5,
                "context_window": 16384,
                "caching": "json",
            },
        },
        "llm_caching": {"json": {"path": "custom_llm.json"}},
    }
    llm_service_overridden = services.instantiate_service(
        "llm", override=override_config
    )
    assert llm_service_overridden.model == "custom_model:1b"
    assert llm_service_overridden.temperature == 0.5
    assert llm_service_overridden.context_window == 16384
    assert llm_service_overridden.service_config is not None
    assert llm_service_overridden.service_config["service_name"] == "llm"


def test_service_params():
    model_name = "abc"
    llm = OllamaClient(model=model_name)
    assert llm._model == model_name
    assert llm.temperature == get_config("llm", "ollama", "temperature")
    # kwargs specified
    model_name = "bcr"
    llm = OllamaClient(model=model_name)
    assert llm._model == model_name

    # using config defaults
    llm = OllamaClient()
    assert llm._model == get_config("llm", "ollama", "model")

    # override
    override_config = {
        "llm": {
            "ollama": {
                "model": "zero",
                "temperature": 0.3,
                "context_window": 8192,
            }
        }
    }
    llm = services.get_service("llm", override=override_config)
    assert llm._model == "zero"
    assert llm.temperature == 0.3


def test_service_singleton():
    services.clear_singletons()
    assert len(services.singletons) == 0

    llm1 = services.get_service("llm")
    llm2 = services.get_service("llm")
    assert llm1.id == llm2.id

    llm3 = services.get_service("llm", variant_name="ollama")
    assert llm1.id == llm3.id
    override_config = {
        "llm": {
            "ollama": {
                "model": "override_model:1b",
                "temperature": 0.3,
                "context_window": 8192,
            }
        }
    }
    llm4 = services.get_service("llm", override=override_config)
    llm5 = services.get_service("llm", override=override_config)
    # for k in services.singletons:

    assert llm4.id == llm5.id
    assert llm4.id != llm1.id



import time
from typing import List
from knwl.llm.llm_cache_base import LLMCacheBase
import ollama

from knwl.llm.llm_base import LLMBase
from knwl.models import KnwlAnswer
from knwl.di import service, inject_config, defaults
from knwl.logging import log


@defaults("@/llm/ollama")
class OllamaClient(LLMBase):
    def __init__(
        self,
        model: str = None,
        temperature: float = None,
        context_window: int = None,
        caching_service: LLMCacheBase = None,
    ):
        super().__init__()
        self.client = (
            ollama.Client()
        )  # the AsyncClient has issues with parallel unit tests and switching models

        self._model = model
        self._temperature = temperature
        self._context_window = context_window

        self._caching_service: LLMCacheBase = caching_service
        self.validate_params()

    def validate_params(self):
        if not self.caching_service:
            log.warn("OllamaClient: No caching service provided, caching disabled.")
        if (
            not isinstance(self.caching_service, LLMCacheBase)
            and self.caching_service is not None
        ):
            raise ValueError(
                f"OllamaClient: caching_service must be an instance of LLMCacheBase, got {type(self.caching_service)}"
            )

    @property
    def model(self):
        return self._model

    @property
    def temperature(self):
        return self._temperature

    @property
    def context_window(self):
        return self._context_window

    @property
    def caching_service(self):
        return self._caching_service

    async def ask(
        self,
        question: str,
        system_message: str = None,
        extra_messages: list[dict] = None,
        key: str = None,
        category: str = None,
    ) -> KnwlAnswer:
        if not question:
            log.warn("OllamaClient: ask called with empty question.")
            return None
        messages = self.assemble_messages(question, system_message, extra_messages)
        if isinstance(messages, str):
            messages = [{"role": "user", "content": messages}]
        # Check cache first
        if self._caching_service is not None:
            cached = await self._caching_service.get(messages, "ollama", self._model)
            if cached is not None:
                return cached
        start_time = time.time()
        response = self.client.chat(
            model=self._model,
            messages=messages,
            options={"temperature": self._temperature, "num_ctx": self._context_window},
        )
        end_time = time.time()
        content = response["message"]["content"]
        answer = KnwlAnswer(
            question=question,
            answer=content,
            messages=messages,
            timing=round(end_time - start_time, 2),
            llm_model=self._model,
            llm_service="ollama",
            key=key if key else question,
            category=category if category else "none",
        )
        if self._caching_service is not None:
            await self._caching_service.upsert(answer)
        return answer

    async def is_cached(self, messages: str | list[str] | list[dict]) -> bool:
        if self.caching_service is None:
            return False
        return await self._caching_service.is_in_cache(messages, "ollama", self._model)

    def __repr__(self):
        return f"<OllamaClient, model={self._model}, temperature={self._temperature},  caching_service={self._caching_service}>"

    def __str__(self):
        return self.__repr__()



# ============================================================================================
# The settings are as much settings as recipe in the case of knwl.
# By enabling/disabling certain features here, you can change the behavior of knwl.
# ============================================================================================
import copy
import os
from typing import Optional

from knwl.utils import get_full_path

"""
Default configuration for Knwl services.
The "default" refers both to the fact that it defines defaults and the "default" space underneath the user's home directory.
That is, you have out of the box a '~/.knwl/user/default' space where all user data is stored unless otherwise specified.
If you use the `Knwl` utility class, you can specify a different space when adding or asking.
"""
_default_config = {"blob": {"default": "file_system", "file_system": {"class": "knwl.storage.file_storage.FileStorage", "base_path": "$/data/blobs", }, }, "chunking": {"default": "tiktoken", "tiktoken": {"class": "knwl.chunking.TiktokenChunking", "model": "gpt-4o-mini", "chunk_size": 1024, "chunk_overlap": 128, }, }, "chunk_store": {"default": "user", "user": {"class": "knwl.semantic.rag.chunk_store.ChunkStore", "chunker": "@/chunking/tiktoken", "chunk_embeddings": "@/vector/user_chunks", "chunk_storage": "@/json/user_chunks", }, }, "document_store": {"default": "user", "user": {"class": "knwl.semantic.rag.document_store.DocumentStore", "document_storage": "@/json/user_documents", }, },
                   "entity_extraction": {"default": "basic", "basic": {"class": "knwl.extraction.basic_entity_extraction.BasicEntityExtraction", "llm": "@/llm", }, }, "graph_extraction": {"default": "full", "full": {"class": "knwl.extraction.basic_graph_extraction.BasicGraphExtraction", "mode": "full",  # fast or full
                                                                                                                                                                                                                        "llm": "@/llm", }, "fast": {"class": "knwl.extraction.basic_graph_extraction.BasicGraphExtraction", "mode": "fast",  # fast or full
                                                                                                                                                                                                                                                    "llm": "@/llm", }, }, "glean_graph_extraction": {"default": "max3", "max3": {"class": "knwl.extraction.glean_graph_extraction.GleanGraphExtraction", "llm": "@/llm", "max_glean": 3, }, }, "graph": {"default": "user", "user": {"class": "knwl.storage.networkx_storage.NetworkXGraphStorage", "format": "graphml", "memory": False, "path": "$/user/default/graph.graphml", }, "memory": {"class": "knwl.storage.networkx_storage.NetworkXGraphStorage", "format": "graphml", "memory": True, }, },
                   "graph_rag": {"default": "user", "memory": {"class": "knwl.semantic.graph_rag.graph_rag.GraphRAG", "semantic_graph": "@/semantic_graph/memory", "ragger": "@/rag_store", "graph_extractor": "@/graph_extraction/basic", "keywords_extractor": "@/keywords_extraction", }, "user": {"class": "knwl.semantic.graph_rag.graph_rag.GraphRAG", "semantic_graph": "@/semantic_graph/user", "ragger": "@/rag_store/user", "graph_extractor": "@/graph_extraction", "keywords_extractor": "@/keywords_extraction", }, },
                   "json": {"default": "basic", "basic": {"class": "knwl.storage.json_storage.JsonStorage", "path": "$/data/data.json", }, "node_store": {"class": "knwl.storage.json_storage.JsonStorage", "path": "$/tests/graphrag/node_store.json", }, "edge_store": {"class": "knwl.storage.json_storage.JsonStorage", "path": "$/tests/graphrag/edge_store.json", }, "document_store": {"class": "knwl.storage.json_storage.JsonStorage", "path": "$/tests/graphrag/document_store.json", }, "chunk_store": {"class": "knwl.storage.json_storage.JsonStorage", "path": "$/tests/graphrag/chunk_store.json", }, "user_documents": {"class": "knwl.storage.json_storage.JsonStorage", "path": "$/user/default/documents.json", },
                            "user_chunks": {"class": "knwl.storage.json_storage.JsonStorage", "path": "$/user/default/chunks.json", }, }, "keywords_extraction": {"default": "basic", "basic": {"class": "knwl.extraction.basic_keywords_extraction.BasicKeywordsExtraction", "llm": "@/llm", }, },
                   "llm": {"default": "ollama", "ollama": {"class": "knwl.llm.ollama.OllamaClient", "model": "qwen2.5:7b", "caching_service": "@/llm_caching/user", "temperature": 0.1, "context_window": 32768, }, "ollama_gemma": {"class": "knwl.llm.ollama.OllamaClient", "model": "gemma3:27b", "caching_service": "@/llm_caching/user", "temperature": 0.1, "context_window": 32768, }, "openai": {"class": "knwl.llm.openai.OpenAIClient", "model": "gpt-4o-mini", "caching_service": "@/llm_caching/user", "temperature": 0.1, "context_window": 32768, }, "anthropic": {"class": "knwl.llm.anthropic.AnthropicClient", "model": "claude-sonnet-4-5-20250929",  # Sonnet 4.5 model
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                 "caching_service": "@/llm_caching/user", "temperature": 0.1, "context_window": 4096,  # Max tokens for response (lower to avoid streaming requirement)
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                 "api_key": os.getenv("ANTHROPY_API_KEY", ""), }, },
                   "llm_caching": {"default": "user", "tests": {"class": "knwl.llm.json_llm_cache.JsonLLMCache", "path": "$/tests/llm.json", }, "user": {"class": "knwl.llm.json_llm_cache.JsonLLMCache", "path": "$/user/default/llm_cache.json", }, }, "logging": {"enabled": True, "level": "INFO", "path": "$/user/default/knwl.log"}, "rag_store": {"default": "user", "user": {"class": "knwl.semantic.rag.rag_store.RagStore", "document_store": "@/document_store/user", "chunk_store": "@/chunk_store/user", "chunker": "@/chunking", "auto_chunk": True, }, }, "semantic_graph": {"default": "user", "user": {"class": "knwl.semantic.graph.semantic_graph.SemanticGraph", "graph_store": "@/graph/user",  # the topology
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                        "node_embeddings": "@/vector/user_nodes",  # the node embeddings
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                        "edge_embeddings": "@/vector/user_edges",  # the edge embeddings
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                        "summarization": "@/summarization",  # how to summarize long texts
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                        }, "memory": {"class": "knwl.semantic.graph.semantic_graph.SemanticGraph", "graph_store": "@/graph/memory",  # the topology
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      "node_embeddings": "@/vector/memory",  # the node embeddings
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      "edge_embeddings": "@/vector/memory",  # the edge embeddings
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      "summarization": "@/summarization",  # how to summarize long texts
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      }, },
                   "summarization": {"default": "llm", "concat": {"class": "knwl.summarization.concat.SimpleConcatenation", "max_tokens": 500, }, "llm": {"class": "knwl.summarization.ollama.OllamaSummarization", "llm": "@/llm", "max_tokens": 150, "chunker": "@/chunking/tiktoken", }, },
                   "vector": {"default": "chroma", "chroma": {"class": "knwl.storage.chroma_storage.ChromaStorage", "memory": False, "path": "$/tests/vector", "collection_name": "default", "metadata": [], }, "user_nodes": {"class": "knwl.storage.chroma_storage.ChromaStorage", "memory": False, "path": "$/user/default/vectors", "collection_name": "nodes", }, "user_edges": {"class": "knwl.storage.chroma_storage.ChromaStorage", "memory": False, "path": "$/user/default/vectors", "collection_name": "edges", }, "user_chunks": {"class": "knwl.storage.chroma_storage.ChromaStorage", "memory": False, "path": "$/user/default/vectors", "collection_name": "chunks", }, "memory": {"class": "knwl.storage.chroma_storage.ChromaStorage", "memory": True, "collection_name": "default", },
                              "chunks": {"class": "knwl.storage.chroma_storage.ChromaStorage", "memory": True, "collection_name": "chunks", }, }, }

_active_config = copy.deepcopy(_default_config)


def set_active_config(new_config: dict):
    """
    Set a new active configuration dictionary.
    This overrides the whole default configuration and can be a better solution then overriding if too many settings need to be changed.

    Args:
        new_config (dict): The new active configuration dictionary to set.
    """
    global _active_config
    _active_config = new_config


def reset_active_config():
    """
    Reset the active configuration to the default configuration.
    """
    global _active_config
    _active_config = copy.deepcopy(_default_config)


def reset_config():
    """
    This is an alias for `reset_active_config` to reset the active configuration to the default configuration.
    """
    reset_active_config()


def merge_configs(override: dict, base_config: dict) -> dict:
    """
    Recursively merge two configuration dictionaries.

    This function merges an override configuration dictionary into a default configuration
    dictionary. For nested dictionaries, the merge is performed recursively. Non-dictionary
    values in the override will replace corresponding values in the default configuration.

    Args:
        override (dict): The configuration dictionary containing values to override defaults.
            Can be None or empty, in which case the base_config is returned unchanged.
        base_config (dict): The base configuration dictionary that will be updated with
            override values. This dictionary is modified in place.

    Returns:
        dict: The merged configuration dictionary (same object as base_config after modification).

    Raises:
        ValueError: If override is not None and not a dictionary.
        ValueError: If base_config is not None and not a dictionary.

    Examples:
        ```python
        default = {
            "a": 1,
            "b": {
                "c": 2,
                "d": 3
            }
        }
        override = {
            "b": {
                "c": 20
            },
            "e": 4
        }
        merged = merge_configs(override, default)
        # merged is now:
        # {
        #     "a": 1,
        #     "b": {
        #         "c": 20,
        #         "d": 3
        #     },
        #     "e": 4
        # }
        ```
    """
    if override is None:
        return base_config
    if not isinstance(override, dict):
        raise ValueError("merge_configs: override must be a dictionary")
    if base_config is None:
        return override
    if not isinstance(base_config, dict):
        raise ValueError("merge_configs: base_config must be a dictionary")

    for key, value in override.items():
        if isinstance(value, dict):
            # get node or create one
            node = base_config.setdefault(key, {})
            merge_configs(value, node)
        else:
            base_config[key] = value
    return base_config


def get_config(*keys, default=None, config=None, override=None):
    """
    Get (recursively) a configuration value from the settings dictionary.
    Args:
        *key: A variable number of string arguments representing the keys to access the nested configuration.
        default: The default value to return if the specified key path does not exist. Defaults to None.
        config: The configuration dictionary to use. If None, the global config will be used. Defaults to None.
        override: An optional dictionary to override the default config for this lookup. Defaults to None.

    Examples:
        ```python
        get_config("llm", "model")
        get_config("llm", "non_existent_key", default="default_value")
        ```
    """
    # the config should not be changed outside
    cloned_config = copy.deepcopy(config or _active_config)
    if len(keys) == 0:
        return cloned_config
    if override is not None:
        cloned_config = merge_configs(override, cloned_config)
    if isinstance(keys[0], str) and keys[0].startswith("@/") and len(keys) > 1:
        # ignore the other keys since things are given via reference
        return get_config(keys[0], default=default, config=cloned_config)
    if len(keys) == 1:
        # if starts with @/, it's a reference to another config value
        if isinstance(keys[0], str) and keys[0].startswith("@/"):
            ref_keys = [u for u in keys[0][2:].split("/") if u]
            if len(ref_keys) == 1:
                if ref_keys[0] not in cloned_config:
                    return default
                # fetch the default variant if only the service name is given
                default_variant = cloned_config.get(ref_keys[0], {}).get("default", None)
                if default_variant is not None:
                    ref_keys.append(default_variant)
                else:
                    raise ValueError(f"get_config: No default variant found for {ref_keys[0]}")
            return get_config(*ref_keys, default=default, config=cloned_config)
        else:
            return cloned_config.get(keys[0], default)
    else:
        current = cloned_config
        # drill down into the nested dictionary
        for k in keys:
            current = current.get(k, None)
            if current is None:
                return default
        return current


def config_exists(*keys, config=None, override=None) -> bool:
    """
    Check if a configuration key path exists in the settings dictionary.

    Args:
        *keys: A variable number of string arguments representing the keys to access the nested configuration.
        config: The configuration dictionary to use. If None, the global config will be used. Defaults to None.
        override: An optional dictionary to override the default config for this lookup. Defaults to None.
    """
    return get_config(*keys, config=config, override=override) is not None


def resolve_dict(d: dict, config: dict = None) -> dict:
    """
    Resolves a configuration dictionary by handling default variants, reference strings, and nested dictionaries.

    If the dictionary contains a "default" key, replaces the dictionary with the variant specified by the "default" value.
    Recursively resolves:
    - Strings starting with "@/": treated as references and resolved via `resolve_reference`.
    - Strings starting with "$/": resolved to full paths via `get_full_path`.
    - Nested dictionaries: resolved recursively.

    Args:
        d (dict): The configuration dictionary to resolve.

    Returns:
        dict: The resolved configuration dictionary.

    Raises:
        ValueError: If the "default" variant specified is not found in the dictionary.
    """
    if "default" in d:
        default_variant = d["default"]
        if default_variant in d:
            d = d[default_variant]
        else:
            raise ValueError(f"resolve_dict: default variant '{default_variant}' not found in configuration.")

    resolved = {}
    for k, v in d.items():
        if isinstance(v, str) and v.startswith("@/"):
            resolved[k] = resolve_reference(v, config=config)
        elif isinstance(v, str) and v.startswith("$/"):
            resolved[k] = get_full_path(v)
        elif isinstance(v, dict):
            resolved[k] = resolve_dict(v, config=config)
        else:
            resolved[k] = v
    return resolved


def resolve_config(*keys, config=None, override=None) -> dict:
    """
    Resolve a configuration for a given service and its default variant.
    You can use '@/' to resolve the whole config if needed.

    Args:
        *keys: A variable number of string arguments representing the keys to access the nested configuration.
        config: The configuration dictionary to use. If None, the global config will be used. Defaults to None.
        override: An optional dictionary to override the default config for this lookup. Defaults to None.
    Returns:
        dict: The resolved configuration dictionary for the specified service and its default variant.
    service_config = get_config(*keys, config=config, override=override)
    if service_config is None:
        raise ValueError(f"resolve_config: No configuration found for {keys}")
    return service_config
    """
    service_config = get_config(*keys, config=config, override=override)
    if service_config is None:
        return None
    if isinstance(service_config, str):
        if service_config.startswith("@/"):
            return resolve_reference(service_config, config=config, override=override)
        else:
            return service_config
    elif isinstance(service_config, dict):
        if "default" in service_config:
            default_variant = service_config["default"]
            if default_variant in service_config:
                service_config = service_config[default_variant]
            else:
                raise ValueError(f"resolve_config: Default variant '{default_variant}' not found in configuration for {keys}")

        resolved = {}
        for k, v in service_config.items():
            if isinstance(v, str) and v.startswith("@/"):
                resolved[k] = resolve_reference(v, config=config, override=override)
            elif isinstance(v, str) and v.startswith("$/"):
                resolved[k] = get_full_path(v)
            else:
                resolved[k] = v
        return resolved
    return service_config


def resolve_reference(ref: str, config=None, override=None) -> dict:
    """
    Resolves iteratively a configuration reference string in the format '@/service/variant'.
    That is, you get a JSON dictionary without any further references inside.

    Args:
        ref: The reference string to resolve (e.g., '@/llm/openai').
        config: The configuration dictionary to use. If None, the global config will be used. Defaults to None.
        override: An optional dictionary to override the default config for this lookup. Defaults to None.
    Returns:
        dict: The resolved configuration dictionary for the specified reference.
    """
    if not ref.startswith("@/"):
        raise ValueError(f"resolve_reference: Invalid reference format: {ref}")
    if ref == "@/":
        # special case to get the whole config
        if override is not None:
            found = merge_configs(override, config or _active_config)
        else:
            found = config or _active_config
    else:
        ref_keys = [u for u in ref[2:].split("/") if u]
        if len(ref_keys) == 1:
            # fetch the default variant if only the service name is given
            default_variant = get_config(ref_keys[0], "default", config=config, override=override)
            if default_variant is not None:
                ref_keys.append(default_variant)
            else:
                return None
        found = get_config(*ref_keys, config=config, override=override)

    # check if any of the values are itself a reference
    if isinstance(found, dict):
        resolved = {}
        for k, v in found.items():
            if isinstance(v, str) and v.startswith("@/"):
                resolved[k] = resolve_reference(v, config=config, override=override)
            elif isinstance(v, str) and v.startswith("$/"):
                resolved[k] = get_full_path(v)
            else:
                resolved[k] = v
        return resolved
    elif isinstance(found, str) and found.startswith("@/"):
        return resolve_reference(found, config=config, override=override)
    if str(found).startswith("$/"):
        return get_full_path(found)
    return found


def get_active_config() -> dict:
    """
    Get a deep copy of the default configuration dictionary.

    Returns:
        dict: A deep copy of the default configuration dictionary.
    """
    return copy.deepcopy(_active_config)


def get_custom_config(namespace: str = "default", llm_provider: str = None, llm_model: str = None, override: Optional[dict] = None, ) -> dict:
    """
    Get the configuration dictionary adjusted for a specific namespace, LLM provider and LLM model.
    - If the namespace is an absolute path (starts with '/'), it will be used as is for storage paths.
    - If no LLM provider or model is given, the default ones from the active config will be used.
    - This is primarily for the Kwnl utility class to create isolated knowledge spaces, but can be used elsewhere too.

    Args:
        namespace: The knowledge space namespace. Defaults to "default". You find this directory under the user's home directory (~/) unless an absolute path is given.
        llm_provider: The LLM provider to set as default (e.g., "openai"). Defaults to None.
        llm_model: The LLM model to set for the given provider (e.g., "gpt-4o-mini"). Defaults to None.
        override: An optional dictionary to override additional configuration settings. Defaults to None.
    """
    if namespace is None or len(str(namespace).strip()) == 0:
        raise ValueError("get_custom_config: namespace cannot be empty.")

    base_config = get_active_config()
    if namespace.startswith("~/"):
        namespace = os.path.expanduser(namespace)
    if namespace.startswith("/"):  # absolute path
        os.makedirs(namespace, exist_ok=True)
        space_path = namespace
    else:
        space_path = f"$/user/{namespace}"

    # replace everywhere the $/user/default with the space path
    def replace_space_path(d):
        for k, v in d.items():
            if isinstance(v, dict):
                replace_space_path(v)
            elif isinstance(v, str) and "$/user/default" in v:
                d[k] = v.replace("$/user/default", space_path)

    # replace the LLM provider and model if given
    def replace_default_provider(d):
        llm_config = d.get("llm", {})
        if llm_provider in llm_config:
            base_config["llm"]["default"] = llm_provider
        else:
            raise ValueError(f"get_custom_config: LLM provider '{llm_provider}' not found in configuration.")

    def replace_default_model(d):
        llm_config = d.get("llm", {})
        provider_config = llm_config.get(llm_provider, {})
        if "model" in provider_config:
            provider_config["model"] = llm_model

        else:
            raise ValueError(f"get_custom_config: LLM model setting not found for provider '{llm_provider}'.")

    replace_space_path(base_config)
    if llm_provider is not None:
        replace_default_provider(base_config)
    if llm_model is not None:
        replace_default_model(base_config)
    # if additional overrides are given, apply them too
    if override is not None:
        base_config = merge_configs(override, base_config)
    return base_config


def merge_into_active_config(section: dict) -> dict:
    """
    Merges an configuration section into the active configuration.
    This a combination of `get_config`, `merge_configs` and `set_active_config`.

    Args:
        section (dict): The configuration dictionary containing values to override in the active configuration.
    """
    current_config = get_config()
    if section is None:
        return current_config
    if not isinstance(section, dict):
        raise ValueError("merge_into_active_config: section must be a dictionary")
    if len(section) == 0:
        return current_config
    new_config = merge_configs(section, current_config)
    set_active_config(new_config)
    return copy.deepcopy(new_config)



import pytest
from faker import Faker
from unittest.mock import patch

from knwl.config import get_config
from knwl.llm.openai import OpenAIClient
from knwl.models.KnwlAnswer import KnwlAnswer
from knwl.utils import get_full_path
from knwl.services import services
pytestmark = pytest.mark.llm

fake = Faker()

@pytest.mark.asyncio
async def test_basic_ask():
   
    llm = OpenAIClient()
    assert llm.model == get_config("llm/openai/model")
    assert llm.temperature == get_config("llm/openai/temperature")

    llm = OpenAIClient(model="gpt-4.1", temperature=0.5)
    assert llm.model == "gpt-4.1"
    assert llm.temperature ==0.5

    # let's change the default caching path
    # note that only the overrides are passed, the rest is taken from default_config
    file_name = fake.word()
    config = {"llm_caching": {"user": {"path": f"$/tests/{file_name}.json"}}}
    llm = services.get_service("llm", "openai", override=config)
    assert llm.caching_service is not None
    assert llm.caching_service.path == get_full_path(f"$/tests/{file_name}.json")
    resp = await llm.ask("Hello")
    assert resp is not None
    assert isinstance(resp, KnwlAnswer)

    assert await llm.is_cached("Hello") is True
    file_path = get_full_path(f"$/tests/{file_name}.json")
    import os

    assert os.path.exists(file_path)
    print("")
    print(resp.answer)


@pytest.mark.asyncio
async def test_override_caching():   

    def create_class_from_dict(name, data):
        return type(name, (), data)

    passed_through_cache = False

    async def is_in_cache(self, *args, **kwargs):
        nonlocal passed_through_cache
        passed_through_cache = True
        return True

    SpecialClass = create_class_from_dict(
        "Special", {"name": "Swa", "is_in_cache": is_in_cache}
    )

    config = {
        "llm": {"openai": {"caching_service": "@/llm_caching/special"}},
        "llm_caching": {"special": {"class": SpecialClass()}},
    }
    with patch(
        "knwl.llm.openai.OpenAIClient.validate_params",
        return_value=None,
    ):    
        llm = services.get_service("llm", "openai", override=config)
        assert llm.caching_service is not None
        assert llm.caching_service.name == "Swa"
        assert await llm.is_cached("Anything") is True
        assert passed_through_cache is True


@pytest.mark.asyncio
async def test_no_cache():
    """
    Test that OpenAI client works correctly when caching is disabled.

    Verifies that:
    - The caching service is None when caching is disabled
    - Messages are not cached when caching is disabled
    - The is_cached method returns False for any message when caching is disabled
    """
    config = {
        "llm": {"openai": {"caching_service": "None"}},
    }
    llm = services.get_service("llm", "openai", override=config)
    await llm.ask("Hello")
    assert await llm.is_cached("Hello") is False



"""
Base classes and interfaces for the Knwl formatting system.

This module provides the foundation for a pluggable, extensible formatting system
that can render Knwl models in various output formats (terminal, HTML, markdown, etc.).
"""

from abc import ABC, abstractmethod
from typing import Any, Dict, Type, Optional, Callable
from pydantic import BaseModel


class FormatterBase(ABC):
    """
    Abstract base class for all formatters.
    
    Formatters are responsible for converting Knwl models into a specific output format.
    Each formatter should implement the format() method to handle the conversion.
    """
    
    @abstractmethod
    def format(self, obj: Any, **options) -> Any:
        """
        Format an object for output.
        
        Args:
            obj: The object to format (typically a Pydantic model)
            **options: Format-specific options
            
        Returns:
            The formatted output (type depends on formatter implementation)
        """
        pass
    
    @abstractmethod
    def render(self, obj: Any, **options) -> None:
        """
        Render an object to the output medium.
        
        Args:
            obj: The object to render
            **options: Rendering options
        """
        pass


class ModelFormatter(ABC):
    """
    Abstract base class for model-specific formatters.
    
    Each Knwl model can have its own formatter that knows how to
    best represent that model in a specific output format.
    """
    
    @abstractmethod
    def format(self, model: BaseModel, formatter: FormatterBase, **options) -> Any:
        """
        Format a specific model instance.
        
        Args:
            model: The Pydantic model instance to format
            formatter: The parent formatter being used (for context/styling)
            **options: Model-specific formatting options
            
        Returns:
            Formatted representation suitable for the parent formatter
        """
        pass


class FormatterRegistry:
    """
    Registry for mapping model types to their formatters.
    
    This allows the system to be easily extended with new models and formatters.
    The registry supports multiple output formats (terminal, html, markdown, etc.)
    """
    
    def __init__(self):
        # Structure: {formatter_type: {model_type: model_formatter}}
        self._formatters: Dict[str, Dict[Type[BaseModel], Type[ModelFormatter]]] = {}
        self._default_formatters: Dict[str, Type[ModelFormatter]] = {}
    
    def register(
        self,
        model_type: Type[BaseModel],
        formatter_class: Type[ModelFormatter],
        format_type: str = "terminal"
    ) -> None:
        """
        Register a model formatter for a specific model and output format.
        
        Args:
            model_type: The Pydantic model class
            formatter_class: The ModelFormatter class to handle this model
            format_type: The output format type (terminal, html, markdown, etc.)
        """
        if format_type not in self._formatters:
            self._formatters[format_type] = {}
        
        self._formatters[format_type][model_type] = formatter_class
    
    def register_default(
        self,
        formatter_class: Type[ModelFormatter],
        format_type: str = "terminal"
    ) -> None:
        """
        Register a default formatter for a format type.
        
        This formatter will be used for any model that doesn't have
        a specific formatter registered.
        
        Args:
            formatter_class: The default ModelFormatter class
            format_type: The output format type
        """
        self._default_formatters[format_type] = formatter_class
    
    def get_formatter(
        self,
        model_type: Type[BaseModel],
        format_type: str = "terminal"
    ) -> Optional[Type[ModelFormatter]]:
        """
        Get the formatter for a specific model and format type.
        
        Args:
            model_type: The Pydantic model class
            format_type: The output format type
            
        Returns:
            The ModelFormatter class, or None if not found
        """
        formatters = self._formatters.get(format_type, {})
        
        # Try exact match first
        if model_type in formatters:
            return formatters[model_type]
        
        # Try base classes
        for base in model_type.__mro__[1:]:
            if base in formatters:
                return formatters[base]
        
        # Fall back to default
        return self._default_formatters.get(format_type)
    
    def has_formatter(
        self,
        model_type: Type[BaseModel],
        format_type: str = "terminal"
    ) -> bool:
        """
        Check if a formatter exists for a model type.
        
        Args:
            model_type: The Pydantic model class
            format_type: The output format type
            
        Returns:
            True if a formatter is registered
        """
        return self.get_formatter(model_type, format_type) is not None


# Global registry instance
_registry = FormatterRegistry()


def register_formatter(
    model_type: Type[BaseModel],
    format_type: str = "terminal"
) -> Callable:
    """
    Decorator to register a model formatter.
    
    Usage:
        @register_formatter(KnwlNode, "terminal")
        class KnwlNodeTerminalFormatter(ModelFormatter):
            ...
    
    Args:
        model_type: The Pydantic model class
        format_type: The output format type
    """
    def decorator(formatter_class: Type[ModelFormatter]) -> Type[ModelFormatter]:
        _registry.register(model_type, formatter_class, format_type)
        return formatter_class
    return decorator


def register_default_formatter(format_type: str = "terminal") -> Callable:
    """
    Decorator to register a default formatter for a format type.
    
    Args:
        format_type: The output format type
    """
    def decorator(formatter_class: Type[ModelFormatter]) -> Type[ModelFormatter]:
        _registry.register_default(formatter_class, format_type)
        return formatter_class
    return decorator


def get_registry() -> FormatterRegistry:
    """Get the global formatter registry."""
    return _registry



# ============================================================================================
# Use VSCode Interactive Python for best experience but you can also run this script directly.
# See https://code.visualstudio.com/docs/python/jupyter-support-py
# ============================================================================================
# %%
from knwl import services

extractor = services.get_service("entity_extraction")
text = """
OpenAI, based in San Francisco, is a leading AI research lab. It was founded in 2015 by Elon Musk and Sam Altman. The company has developed several groundbreaking AI models, including GPT-3 and DALL-E. Microsoft has invested heavily in OpenAI, integrating its technology into products like Azure and Office 365. Other notable AI companies include Google DeepMind, known for AlphaGo, and Anthropic, which focuses on AI safety.
"""
entities = ["PRODUCT"]
results = await extractor.extract(text, entities)
for record in results:
    print(record.model_dump_json(indent=2))

# %%
# ============================================================================================
# Extract and specify multiple entity types.
# ============================================================================================

entities = ["ORG", "PERSON", "GPE"]
results = await extractor.extract(text, entities)
for record in results:
    print(record.model_dump_json(indent=2))

# %%



import os
import uuid
from abc import ABC
from pathlib import Path

from knwl.config import get_config


class FrameworkBase(ABC):
    """
    Base class for all framework components providing common utilities.

    The `id` attribute is a unique identifier for each instance and is assigned for convenience.

    """
    def __init__(self, *args, **kwargs):
       
        self.id = str(uuid.uuid4())


    def get_service(self, service_name: str, variant_name: str = None, override=None):
        from knwl.services import services
        return services.get_service(
            service_name, variant_name=variant_name, override=override
        )

    def get_llm(self, llm_variant: str = None, override=None):
        return self.get_service("llm", llm_variant, override=override)

    def get_user_home(self) -> str:
        home = Path.home()
        if home is None:
            raise ValueError("Home directory not found")
        return str(home)

    def ensure_path_exists(self, path: str) -> str:
        if path is None:
            raise ValueError("Path cannot be None")

        # determine whether the path is an absolute path or relative path
        if not Path(path).is_absolute():
            path = os.path.join(self.get_user_home(), path)
        # if the path is a file path and not a directory, get the parent directory
        if os.path.isfile(path):
            path = os.path.dirname(path)
        if not os.path.exists(path):
            os.makedirs(path, exist_ok=True)
        return path

    @staticmethod
    def get_test_dir():
        # Get the directory of the current file (framework_base.py) to ensure consistent path resolution
        current_dir = os.path.dirname(__file__)
        return os.path.join(current_dir, "..", "tests", "data")

    def get_param(self, keys, args=None, kwargs=None, default=None, override=None):
        """
        Utility to get parameter from kwargs, args or config in that order.
        The `override` parameter can be used to provide an override configuration dictionary, otherwise the default config will be used.

        Args:
        keys: list of keys to drill down in config
        args: positional arguments (list)
        kwargs: keyword arguments (dict)
        default: default value if not found
        override: override config dict
        """
        if len(keys) == 0:
            return None
        key = keys[-1]
        if kwargs and key in kwargs:
            return kwargs.get(key)
        if args:
            for arg in args:
                if isinstance(arg, dict) and key in arg:
                    return arg.get(key)
        return get_config(*keys, default=default, override=override)



from typing import Optional
from pydantic import BaseModel, Field, model_validator

from knwl.utils import hash_with_prefix


class KnwlReference(BaseModel):
    """
    Represents a reference in the RAG (Retrieval-Augmented Generation) system.
    This class holds metadata about a reference, including its index, name, description,
    timestamp, and a unique identifier (id).
    It is used to track and manage references within the context of RAG operations.

    Chunks refer to specific pieces of information or data that are retrieved
    during the RAG process, and references provide additional context or metadata
    about these chunks.
    The `id` is typically a hash of the reference's content, ensuring uniqueness.
    
    Attributes:
        index (str): The index identifier of the reference.
        name (str): The name of the reference.
        description (str): A description of the reference.
        timestamp (str): The timestamp when the reference was created or last modified.
        id (str): The unique identifier of the reference, typically a hash of its content.
    """
    
    index: Optional[int] = Field(default=0,description="The index within the list of references.")
    document_id: str = Field(description="The unique identifier of the document containing the reference.")
    content: Optional[str] = Field(default=None, description="The content of the reference.")
    name: Optional[str] = Field(default=None, description="The name of the reference.")
    description: Optional[str] = Field(default=None, description="A description of the reference.")
    timestamp: Optional[str] = Field(default=None, description="The timestamp when the reference was created or last modified.")
    id: Optional[str] = Field(default=None, description="The unique identifier of the reference, typically a hash of its content.")

    @model_validator(mode="after")
    def set_id(self) -> "KnwlReference":
        if self.content is not None and len(str.strip(self.content)) > 0:
            object.__setattr__(self, "id", KnwlReference.hash_keys(self.content))
        return self

    @staticmethod
    def hash_keys(content: str) -> str:
        return hash_with_prefix(content, prefix="grag-ref|>")

    


from abc import ABC, abstractmethod

from knwl.framework_base import FrameworkBase
from knwl.models.KnwlChunk import KnwlChunk
from knwl.models.KnwlDocument import KnwlDocument


class RagBase(FrameworkBase, ABC):
    """
    Base class to manage documents and chunks.
    The default implementation is `RagStore`, which uses separate storage for documents and chunks. Of course, nothing prevents you to use the same storage for both.
    """

    @abstractmethod
    async def upsert_document(self, obj: str | KnwlDocument) -> str:
        """
        Upserts a document into the document store.

        Args:
            obj (str): The document text to be upserted.

        Returns:
            str: The ID of the upserted document.
        """
        ...

    @abstractmethod
    async def get_document_by_id(self, document_id: str) -> KnwlDocument | None:
        """
        Retrieves a document by its ID.

        Args:
            document_id (str): The ID of the document to retrieve.

        Returns:
            KnwlDocument | None: The retrieved document or None if not found.
        """
        ...
        ...

    @abstractmethod
    async def delete_document_by_id(self, document_id: str) -> None:
        """
        Deletes a document by its ID.

        Args:
            document_id (str): The ID of the document to delete.
        """
        ...

    @abstractmethod
    async def upsert_chunk(self, obj: str | KnwlChunk) -> str:
        """
        Upserts a chunk into the chunk store.

        Args:
            obj (str): The chunk text to be upserted.

        Returns:
            str: The ID of the upserted chunk.
        """
        ...

    @abstractmethod
    async def get_chunk_by_id(self, chunk_id: str) -> KnwlChunk | None:
        """
        Retrieves a chunk by its ID.

        Args:
            chunk_id (str): The ID of the chunk to retrieve.

        Returns:
            KnwlChunk | None: The retrieved chunk or None if not found.
        """
        ...

    @abstractmethod
    async def delete_chunk_by_id(self, chunk_id: str) -> None:
        """
        Deletes a chunk by its ID.

        Args:
            chunk_id (str): The ID of the chunk to delete.
        """
        ...

    @abstractmethod
    async def nearest(self, query: str, top_k: int = 5) -> list[KnwlChunk]:
        """
        Retrieves the nearest chunks based on a query.

        Args:
            query (str): The query string to search for.
            top_k (int): The number of top results to return.

        Returns:
            list[KnwlChunk]: A list of the nearest chunks.
        """
        ...

    @abstractmethod
    async def chunk(self, document: KnwlDocument) -> list[KnwlChunk]:
        ...



from abc import ABC, abstractmethod
from typing import List

from knwl.framework_base import FrameworkBase
from knwl.models import (
    KnwlParams,
    KnwlChunk,
    KnwlContext,
    KnwlGraph,
    KnwlInput,
    KnwlInput,
)
from knwl.models.KnwlDocument import KnwlDocument
from knwl.models.KnwlEdge import KnwlEdge
from knwl.models.KnwlIngestion import KnwlIngestion
from knwl.models.KnwlKeywords import KnwlKeywords
from knwl.models.KnwlNode import KnwlNode


class GraphRAGBase(FrameworkBase, ABC):
    @abstractmethod
    async def embed_node(self, node: KnwlNode) -> KnwlNode | None: ...

    @abstractmethod
    async def embed_nodes(self, nodes: list[KnwlNode]) -> list[KnwlNode]: ...

    @abstractmethod
    async def embed_edge(self, edge: KnwlEdge) -> KnwlEdge | None: ...

    @abstractmethod
    async def embed_edges(self, edges: list[KnwlEdge]) -> list[KnwlEdge]: ...

    @abstractmethod
    async def ingest(self, input: str | KnwlInput | KnwlDocument) -> KnwlGraph | None:
        """
        Ingest raw text or KnwlInput/KnwlDocument and convert to knowledge graph:
        - Chunk the text if necessary
        - Extract entities and relationships
        - Embed (consolidate) nodes and edges (graph and vector store)
        """
        ...

    @abstractmethod
    async def extract(
        self, input: str | KnwlInput | KnwlDocument, enable_chunking: bool = True
    ) -> KnwlIngestion | None:
        """
        Extract a knowledge graph from raw text or KnwlInput/KnwlDocument.
        This is the same as `ingest` but without embedding (consolidation).
        """
        ...

    @abstractmethod
    async def extract_keywords(self, input: str | KnwlInput) -> KnwlKeywords | None:
        """
        Extract keywords from the input text.
        These keywords will be used to query the knowledge graph for relevant context.
        The high-level keywords are global topics, main subjects or areas of interest in the text.
        The local keywords are specific entities, concepts, or details mentioned in the text.
        """
        ...

    @abstractmethod
    async def augment(self, input: str | KnwlInput) -> KnwlContext | None:
        """
        Retrieve context from the knowledge graph and augment the input text.
        All you need to answer questions or generate text with context.
        """
        ...

    @abstractmethod
    async def nearest_nodes(
        self, query: str, query_param: KnwlParams
    ) -> list[KnwlNode] | None:
        """
        Query nodes from the knowledge graph based on the input query and parameters.
        """
        ...

    @abstractmethod
    async def nearest_edges(
        self, query: str, params: KnwlParams
    ) -> list[KnwlEdge] | None:
        """
        Query edges from the knowledge graph based on the input query and parameters.
        """
        ...

    @abstractmethod
    async def nearest_chunks(
        self, query: str, params: KnwlParams
    ) -> list[KnwlChunk] | None:
        """
        Query chunks based on the input query and parameters.
        This does not involve the graph directly but is part of the naive RAG pipeline.
        """
        ...

    @abstractmethod
    async def get_node_by_id(self, node_id: str) -> KnwlNode | None:
        """
        Retrieve a node by its ID from the knowledge graph.
        """
        ...

    @abstractmethod
    async def node_degree(self, node_id: str) -> int:
        """
        Retrieve the degree of a node in the graph.

        Args:
            node_id (str): The unique identifier of the node.

        Returns:
            int: The degree of the node, or 0 if the node does not exist.
        """
        ...

    @abstractmethod
    async def get_attached_edges(self, nodes: list[KnwlNode]) -> list[KnwlEdge]:
        """
        Retrieve the edges attached to the given nodes.

        Args:
            nodes (list[KnwlNode]): A list of KnwlNode objects for which to retrieve attached edges.

        Returns:
            list[KnwlEdge]: A list of KnwlEdge objects attached to the given nodes.
        """
        ...

    @abstractmethod
    async def get_edges_between_nodes(
        self, source_id: str, target_id: str
    ) -> list[KnwlEdge]:
        """
        Retrieve edges between two nodes by their IDs from the knowledge graph.
        """
        ...

    @abstractmethod
    async def save_sources(self, sources: list[KnwlDocument]) -> bool:
        """
        Save the source documents used for ingestion.
        This is important for traceability and reference but implementation is optional if this operates within a broader system that already manages documents (workflow).

        Args:
            sources (list[KnwlDocument]): A list of KnwlDocument objects representing the source documents.

        Returns:
            bool: True if the sources were saved successfully, False otherwise.
        """
        ...

    @abstractmethod
    async def save_chunks(self, chunks: list[KnwlChunk]) -> bool:
        """
        Save the chunks generated during ingestion.
        This is important for traceability and reference but implementation is optional if this operates within a broader system that already manages text chunks (workflow).

        Args:
            chunks (list[KnwlChunk]): A list of KnwlChunk objects representing the text chunks.

        Returns:
            bool: True if the chunks were saved successfully, False otherwise.
        """
        ...

    @abstractmethod
    async def get_chunk_by_id(self, chunk_id: str) -> KnwlChunk | None:
        """
        Retrieve a chunk by its Id from the chunk storage.
        The implementation of this method is optional depending on whether chunk storage is managed within this system or externally.
        Args:
            chunk_id (str): The unique identifier of the chunk.
        """
        ...

    @abstractmethod
    async def get_source_by_id(self, source_id: str) -> KnwlDocument | None:
        """
        Retrieve a source document by its Id from the source storage.
        The implementation of this method is optional depending on whether source storage is managed within this system or externally.
        """
        ...

    @abstractmethod
    async def edge_degree(self, edge_id: str) -> int: ...

    @abstractmethod
    async def assign_edge_degrees(self, edges: list[KnwlEdge]) -> list[int]: ...

    @abstractmethod
    async def get_semantic_endpoints(
        self, edge_ids: list[str]
    ) -> dict[str, tuple[str, str]]: ...

    @abstractmethod
    async def node_exists(self, node_id: str) -> bool:
        """
        Check if a node with the given Id exists in the knowledge graph.
        """
        ...

    @abstractmethod
    async def node_count(self) -> int:
        """
        Get the total number of nodes in the knowledge graph.
        """
        ...

    @abstractmethod
    async def edge_count(self) -> int:
        """
        Get the total number of edges in the knowledge graph.
        """
        ...

    @abstractmethod
    async def delete_node_by_id(self, node_id: str) -> bool:
        """
        Delete a node by its Id from the knowledge graph.
        Returns True if the node was deleted, False if it did not exist.
        """
        ...

    @abstractmethod
    async def chunk(self, document: KnwlDocument) -> list[KnwlChunk]:
        """
        Chunk the given document into smaller KnwlChunk objects.
        """
        ...



from abc import ABC, abstractmethod

from knwl.models import KnwlBlob
from knwl.storage.storage_base import StorageBase


class BlobStorageBase(StorageBase, ABC):
    """
    Abstract base class for storing blobs.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @abstractmethod
    async def upsert(self, blob: KnwlBlob) -> str | None: ...

    @abstractmethod
    async def get_by_id(self, id:str) -> KnwlBlob | None: ...

    @abstractmethod
    async def get_by_id(self, ids, fields=None): ...

    @abstractmethod
    async def delete_by_id(self, id: str) -> bool: ...

    @abstractmethod
    async def count(self) -> int: ...

    @abstractmethod
    async def exists(self, id: str) -> bool: ...



from knwl.models import *
from knwl.services import services
from knwl.semantic.graph_rag.graph_rag import GraphRAG
from knwl.format import print_knwl
from knwl.chunking.chunking_base import ChunkingBase
from knwl.chunking.tiktoken_chunking import TiktokenChunking
from knwl.di import (
    singleton_service,
    inject_config,
    inject_services,
    auto_inject,
    defaults,
)
from knwl.prompts import prompts
from knwl.knwl import Knwl



import pytest

from knwl.format import print_knwl
from knwl.models import KnwlInput, KnwlParams
from knwl.semantic.graph_rag.strategies.self_strategy import SelfGragStrategy
pytestmark = pytest.mark.llm


@pytest.mark.asyncio
async def test_self_strategy_augment():
    question = "What is the significance of DNA in living organisms?"
    strategy = SelfGragStrategy(grag=None)  # this strategy does not use grag in augment
    input = KnwlInput(text=question, params=KnwlParams(strategy="self", return_chunks=True))
    found = await strategy.augment(input)
    print("")
    print_knwl(found, show_texts=True, show_nodes=False, show_edges=False)



from abc import ABC, abstractmethod
from typing import Optional

from knwl.framework_base import FrameworkBase
from knwl.models.KnwlDocument import KnwlDocument
from knwl.storage.storage_base import StorageBase


class DocumentBase(FrameworkBase, ABC):
    def __init__(
        self,
        document_storage: Optional[StorageBase] = None,
    ):
        super().__init__()
        self.document_storage: StorageBase = document_storage

    @abstractmethod
    async def upsert(self, obj: str | KnwlDocument) -> str: ...

    @abstractmethod
    async def get_by_id(self, document_id: str) -> KnwlDocument | None: ...

    @abstractmethod
    async def delete_by_id(self, document_id: str) -> None: ...

    @abstractmethod
    async def exists(self, document_id: str) -> bool: ...





import logging
import os
import tempfile
from pathlib import Path
from unittest import mock
import pytest
from knwl.logging import Log

class TestLog:
    """Test suite for Log class"""

    def setup_method(self):
        """Setup for each test - clear handlers"""
        logger = logging.getLogger("knwl")
        for handler in logger.handlers[:]:
            handler.close()
            logger.removeHandler(handler)

    def teardown_method(self):
        """Cleanup after each test"""
        logger = logging.getLogger("knwl")
        for handler in logger.handlers[:]:
            handler.close()
            logger.removeHandler(handler)

    def test_log_initialization_default(self):
        """Test Log initialization with default settings"""
        log = Log()
        assert log.enabled is True
        assert log.logging_level == "INFO"
        assert log.logger is not None
        assert log.logger.name == "knwl"
        log.shutdown()

    def test_log_initialization_with_override(self):
        """Test Log initialization with config override"""
        override = {
            "logging": {
                "enabled": True,
                "level": "DEBUG",
                "path": "$/user/test/test.log"
            }
        }
        log = Log(override=override)
        assert log.enabled is True
        assert log.logging_level == "DEBUG"
        assert "test.log" in log.path
        log.shutdown()

    def test_log_disabled(self):
        """Test Log when logging is disabled"""
        override = {"logging": {"enabled": False}}
        log = Log(override=override)
        assert log.enabled is False
        assert log.logger is None
        log.shutdown()

    def test_log_levels(self):
        """Test different logging levels"""
        levels = ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]
        for level in levels:
            override = {"logging": {"level": level}}
            log = Log(override=override)
            assert log.logging_level == level
            log.shutdown()

    def test_invalid_log_level(self):
        """Test that invalid log level raises ValueError"""
        override = {"logging": {"level": "INVALID"}}
        with pytest.raises(ValueError, match="Invalid LOGGING_LEVEL"):
            Log(override=override)

    def test_info_logging(self, capsys):
        """Test info level logging"""
        override = {"logging": {"enabled": False}}
        log = Log(override=override)
        log.info("Test info message")
        captured = capsys.readouterr()
        assert "INFO: Test info message" in captured.out
        log.shutdown()

    def test_error_logging(self, capsys):
        """Test error level logging"""
        override = {"logging": {"enabled": False}}
        log = Log(override=override)
        log.error("Test error message")
        captured = capsys.readouterr()
        assert "ERROR: Test error message" in captured.out
        log.shutdown()

    def test_warning_logging(self, capsys):
        """Test warning level logging"""
        override = {"logging": {"enabled": False}}
        log = Log(override=override)
        log.warning("Test warning message")
        captured = capsys.readouterr()
        assert "WARNING: Test warning message" in captured.out
        log.shutdown()

    def test_warn_logging(self, capsys):
        """Test warn (alias for warning) level logging"""
        override = {"logging": {"enabled": False}}
        log = Log(override=override)
        log.warn("Test warn message")
        captured = capsys.readouterr()
        assert "WARNING: Test warn message" in captured.out
        log.shutdown()

    def test_debug_logging(self, capsys):
        """Test debug level logging"""
        override = {"logging": {"enabled": False}}
        log = Log(override=override)
        log.debug("Test debug message")
        captured = capsys.readouterr()
        assert "DEBUG: Test debug message" in captured.out
        log.shutdown()

    def test_exception_logging(self, capsys):
        """Test exception logging"""
        override = {"logging": {"enabled": False}}
        log = Log(override=override)
        try:
            raise ValueError("Test exception")
        except Exception as e:
            log.exception(e)
        captured = capsys.readouterr()
        assert "EXCEPTION: Test exception" in captured.out
        log.shutdown()

    def test_callable_with_message(self, capsys):
        """Test calling Log instance with a message"""
        override = {"logging": {"enabled": False}}
        log = Log(override=override)
        log("Test callable message")
        captured = capsys.readouterr()
        assert "INFO: Test callable message" in captured.out
        log.shutdown()

    def test_callable_with_exception(self, capsys):
        """Test calling Log instance with an exception"""
        override = {"logging": {"enabled": False}}
        log = Log(override=override)
        try:
            raise RuntimeError("Test runtime error")
        except Exception as e:
            log(e)
        captured = capsys.readouterr()
        assert "EXCEPTION: Test runtime error" in captured.out
        log.shutdown()

    def test_callable_without_args(self):
        """Test calling Log instance without arguments raises ValueError"""
        log = Log()
        with pytest.raises(ValueError, match="You can only call the log directly"):
            log()
        log.shutdown()

    def test_file_logging_creation(self):
        """Test that file logging creates log file"""
        with tempfile.TemporaryDirectory() as tmpdir:
            log_path = Path(tmpdir) / "test.log"
            override = {"logging": {"path": str(log_path)}}
            log = Log(override=override)
            log.info("Test message")
            log.shutdown()
            assert log_path.exists()

    def test_no_duplicate_handlers(self):
        """Test that handlers are not duplicated"""
        log1 = Log()
        initial_handler_count = len(log1.logger.handlers)
        log1.shutdown()
        
        log2 = Log()
        assert len(log2.logger.handlers) == initial_handler_count
        log2.shutdown()

    def test_shutdown(self, capsys):
        """Test shutdown closes and removes handlers"""
        log = Log()
        log.shutdown()
        captured = capsys.readouterr()
        assert "Logging system shut down successfully" in captured.out
        assert len(log.logger.handlers) == 0

    def test_file_handler_rotation_config(self):
        """Test that file handler is configured with rotation settings"""
        with tempfile.TemporaryDirectory() as tmpdir:
            log_path = Path(tmpdir) / "test.log"
            override = {"logging": {"path": str(log_path)}}
            log = Log(override=override)
            
            # Find the rotating file handler
            file_handler = None
            for handler in log.logger.handlers:
                if isinstance(handler, logging.handlers.RotatingFileHandler):
                    file_handler = handler
                    break
            
            assert file_handler is not None
            assert file_handler.maxBytes == 10 * 1024 * 1024  # 10MB
            assert file_handler.backupCount == 5
            log.shutdown()

    def test_console_handler_exists(self):
        """Test that console handler is added"""
        log = Log()
        
        has_console_handler = any(
            isinstance(handler, logging.StreamHandler)
            for handler in log.logger.handlers
        )
        
        assert has_console_handler is True
        log.shutdown()

    @mock.patch('knwl.logging.RotatingFileHandler')
    def test_file_logging_failure_handling(self, mock_handler, capsys):
        """Test graceful handling of file logging setup failure"""
        mock_handler.side_effect = Exception("File creation failed")
        
        log = Log()
        captured = capsys.readouterr()
        assert "Failed to set up file logging" in captured.out
        log.shutdown()

    @mock.patch('logging.StreamHandler')
    def test_console_logging_failure_handling(self, mock_handler, capsys):
        """Test graceful handling of console logging setup failure"""
        mock_handler.side_effect = Exception("Console handler failed")
        
        log = Log()
        captured = capsys.readouterr()
        assert "Failed to set up console logging" in captured.out
        log.shutdown()


import pytest

from knwl.storage.file_storage import FileStorage
from knwl.models import KnwlBlob
import os
import shutil
pytestmark = pytest.mark.basic


@pytest.mark.asyncio
async def test_file_storage_upsert_get_delete():
    # Setup
    storage = FileStorage(base_path="$/tests/file_storage")
    blob = KnwlBlob(
        id="test_blob",
        data=b"Much ado about nothing.",
        name="Test Blob",
        description="A blob for testing and experimentation.",
        metadata={"author": "Shakespeare", "year": 1600},
    )

    # Test upsert
    blob_id = await storage.upsert(blob)
    assert blob_id == blob.id

    # Test get_by_id
    retrieved_blob = await storage.get_by_id("test_blob")
    assert retrieved_blob is not None
    assert retrieved_blob.id == "test_blob"
    assert retrieved_blob.data == b"Much ado about nothing."
    assert retrieved_blob.name == "Test Blob"
    assert retrieved_blob.description == "A blob for testing and experimentation."
    assert retrieved_blob.metadata == {"author": "Shakespeare", "year": 1600}

    # Test exists
    exists = await storage.exists("test_blob")
    assert exists is True

    # Test count
    count = await storage.count()
    assert count == 1

    # Test delete_by_id
    deleted = await storage.delete_by_id("test_blob")
    assert deleted is True

    # Verify deletion
    retrieved_blob_after_delete = await storage.get_by_id("test_blob")
    assert retrieved_blob_after_delete is None

    # Cleanup
    shutil.rmtree(storage.base_path, ignore_errors=True)



from abc import ABC, abstractmethod
from typing import Any

from knwl.storage.storage_base import StorageBase


class KeyValueStorageBase(StorageBase, ABC):
    """
    Abstract base class for JSON dictionary storage on disk.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @abstractmethod
    async def get_all_ids(self) -> list[str]:
        pass

    @abstractmethod
    async def save(self):
        pass

    @abstractmethod
    async def clear_cache(self):
        pass

    @abstractmethod
    async def get_by_id(self, id):
        pass

    @abstractmethod
    async def get_by_ids(self, ids, fields=None):
        pass

    @abstractmethod
    async def filter_new_ids(self, data: list[str]) -> set[str]:
        pass

    @abstractmethod
    async def upsert(self, obj: Any):
        pass

    @abstractmethod
    async def clear(self):
        pass

    @abstractmethod
    async def count(self):
        pass

    @abstractmethod
    async def delete_by_id(self, id: str):
        pass
    @abstractmethod
    async def exists(self, id: str) -> bool:
        pass


# ============================================================================================
# Use VSCode Interactive Python for best experience but you can also run this script directly.
# See https://code.visualstudio.com/docs/python/jupyter-support-py
# ============================================================================================
# %%
"""
==============================================================================================
Knwl comes with a predefined set of default configurations that allow you to get started. You don't need to install anything beyond the package dependencies (ie. `uv sync`).

Of course, somewhere down the line you need a LLM service and the default is OpenAI. As such, you need to set the `OPENAI_API_KEY` environment variable to your OpenAI API key. If you have Ollama installed you can also use Ollama, see below.

The `Kwnl` class is a utility class that wraps various functionalities without having to instantiate or configure anything. It's useful for quick experiments and prototyping, but the full power of Knwl is unleashed when you start configuring your own services, spaces, and strategies.
==============================================================================================
"""
from knwl import Knwl, print_knwl

knwl = Knwl()
# %%
"""
==============================================================================================
You can ask questions directly and this uses the default LLM configured.
==============================================================================================
"""
a = await knwl.ask("DNA is the essence of life.")
print_knwl(a)  # pretty print the KnwlAnswer
print(a.answer)  # just print the answer string

"""
The above prints something like:
╭───────────────────────────────── KnwlAnswer ─────────────────────────────────╮
│                                                                              │
│   messages      [1 items]                                                    │
│   llm_model     gpt-oss:20b                                                  │
│   llm_service   ollama                                                       │
│   answer        Absolutely—DNA is the blueprint that carries the             │
│                 instructions for building and maintaining every living       │
│                 organism. Its double‑helix structure stores...               │
│   timing        7.95                                                         │
│   key           DNA is the essence of life.                                  │
│   category      none                                                         │
│   question      DNA is the essence of life.                                  │
│   from_cache    False                                                        │
│   id            answer|>0a93fcc2ba54d2c33155dcdb4893f0d6                     │
│                                                                              │
╰──────────────────────────────────────────────────────────────────────────────╯
"""

# %%
"""
==============================================================================================
How was this question answered? You can inspect the configuration used:
- @/llm shows the default LLM configuration
- the LLM service can have multiple models and variations
==============================================================================================
"""
print_knwl("@/llm")

# %%
"""
==============================================================================================
You can add single facts to the knowledge graph using the `add_fact` method.
There is no need to specify the id since this is a hash of the name the type.
The type is also optional and defaults to "Unknown".
==============================================================================================
"""
await knwl.add_fact(
    "gravity",
    "Gravity is a universal force that attracts two bodies toward each other.",
    type="Fact",
)

await knwl.add_fact(
    "photosynthesis",
    "Photosynthesis is the process by which green plants and some other organisms use sunlight to synthesize foods from carbon dioxide and water.",
    type="Fact",
)
await knwl.connect(
    source_name="gravity",
    target_name="photosynthesis",
    relation="Both are fundamental natural processes.",
)
# %%
"""
==============================================================================================
Of course, you typically do not create a KG manually like this.
You can add arbitrary text and ingest it using the `add` method.
==============================================================================================
"""

result = await knwl.add(
    "Quantum topology is all about knot theory and how it applies to quantum physics. Whether it's also underpinning quantum gravity remains to be seen, but it's a fascinating area of scientific exploration.",
)
# %%
print_knwl(result)
"""
╭───────────────────────────── 👁️ Knowledge Graph ──────────────────────────────╮
│ Id: bf3edc7e-8370-4525-a92f-52fde401efaf                                     │
│ Nodes: 2, Edges: 1                                                           │
│ Keywords: knot theory, scientific exploration...                             │
│                                                                              │
│                                                                              │
│ 🔵 Nodes:                                                                    │
│ Quantum Topology : concept - Quantum Topology is a field of study focusing   │
│ on the application of knot theory in quantum physics, exploring the          │
│ mathematical aspects of quantum mech...                                      │
│ Knot Theory : concept - Knot Theory is a branch of mathematics that studies  │
│ the properties of knots, which can be applied to various areas, including    │
│ quantum physics, to unde...                                                  │
│                                                                              │
│                                                                              │
│ 🔗 Edges:                                                                    │
│ node|>7a ─[mathematics]→ node|>66                                            │
╰────────────────────────────── 2 nodes, 1 edges ──────────────────────────────╯
"""


# %%
"""
==============================================================================================
Kwnl allows you to use namespaces for different knowledge spaces.
==============================================================================================
"""
knwl = Knwl("geography")
a = await knwl.ask("What is the capital of Tanzania?")
print_knwl(a)

# %%
"""
==============================================================================================
You can also specify different LLM providers and models.
==============================================================================================
"""
knwl = Knwl("swa", llm="openai", model="gpt-5-nano")
a = await knwl.ask("What is the capital of Tanzania?")
print_knwl(a)
# %%
"""
==============================================================================================
Classic RAG does not find information that is not vector-similar to the question.
However, Knwl's graph RAG can find related information via graph connections.
Below we create two nodes and connect them. Despite that "gravity" is not related to "photosynthesis" in a vector space,
the graph connection allows Knwl to find and use the gravity fact when answering a question about photosynthesis.
This is the most basic (but powerful) demonstration how graph RAG can improve over classic RAG.
==============================================================================================
"""

knwl = Knwl()

# add a fact
await knwl.add_fact("gravity", "Gravity is a universal force that attracts two bodies toward each other.", id="fact1", )

assert (await knwl.node_exists("fact1")) is True

# add another fact
await knwl.add_fact("photosynthesis", "Photosynthesis is the process by which green plants and some other organisms use sunlight to synthesize foods from carbon dioxide and water.", id="fact2", )

# you can take the node returned from add_fact as an alternative
found = await knwl.get_nodes_by_name("gravity")
gravity_node = found[0]
found = await knwl.get_nodes_by_name("photosynthesis")
photosynthesis_node = found[0]
# connect the two nodes
await knwl.connect(source_name=gravity_node.name, target_name=photosynthesis_node.name, relation="Both are fundamental natural processes.", )

# Augmentation will fetch the gravity node, despite that it does not directly relate to photosynthesis
# Obviously, this 1-hop result would not happen with classic RAG since the vector similarity is too low
augmentation = await knwl.augment("What is photosynthesis?")
# pretty print the augmentation result
print_knwl(augmentation)

# graph RAG question-answer
a = await knwl.ask("What is photosynthesis?")
print_knwl(a.answer)
# %%
